import matplotlib

gui_env = [i for i in matplotlib.rcsetup.interactive_bk]
non_gui_backends = matplotlib.rcsetup.non_interactive_bk
print("Non Gui backends are:", non_gui_backends)
print("Gui backends are: ", gui_env)
for gui in gui_env + non_gui_backends:  # ['agg', 'tkAgg']:
    try:
        matplotlib.use(gui, force=True)
        from matplotlib import pyplot as plt

        plt.plot([1.5, 2.0, 2.5])
        fig = plt.gcf()
        fig.suptitle(gui)
        # plt.show()
        print(f"OK     {gui.rjust(10)} :" + matplotlib.get_backend())
    except Exception as _excp:
        print(f"FAILED {gui.rjust(10)} : {_excp}")



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii

__all__ = ['OMFITkeplerParams']


class OMFITkeplerParams(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface with kepler input files

    :param filename: filename passed to OMFITascii class

    :param \**kw: keyword dictionary passed to OMFITascii class
    """

    def __init__(self, filename, **kw):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.read().split('\n')
        for line in lines:
            if '=' not in line:
                continue
            key, value = line.split('=')
            key = key.strip()
            value = value.strip()
            if value == 'false':
                value = False
            elif value == 'true':
                value = True
            elif value == 'none':
                value = None
            else:
                try:
                    value = ast.literal_eval(value)
                except (ValueError, SyntaxError):
                    if '\\,' in value:
                        value = list(map(lambda x: ast.literal_eval(x.strip()), value.split('\\,')))
                    else:
                        value = re.sub(r'\\', '', value)
            key = re.sub(r'\\', '', key)
            h = self
            for k in key.split('.')[:-1]:
                if k not in h:
                    h[k] = SortedDict()
                h = h[k]
            h[key.split('.')[-1]] = value

    @dynaSave
    def save(self):
        def escape(inv):
            if isinstance(inv, list):
                return '\\,'.join(map(str, inv))
            elif inv in [True, False, None]:
                return str(inv).lower()
            elif not isinstance(inv, str):
                return inv
            return re.sub('([ ,])', r'\\\1', inv)

        with open(self.filename, 'w') as f:
            for key in traverse(self, onlyLeaf=True):
                value = eval('self' + key)
                key = '.'.join(parseLocation(key))[1:]
                f.write('%s = %s\n' % (escape(key), escape(value)))


############################################
if '__main__' == __name__:
    test_classes_main_header()

    tmp = OMFITkeplerParams(OMFITsrc + '/../samples/kepler_input_sample.txt')
    tmp.load()
    tmp.saveas(os.getcwd() + 'kepler_input_sample.txt')



from pprint import pformat

if compare_version(omas.__version__, '0.70') < 0:
    raise Exception('machine to OMAS mappings is supported only after version 0.70.0 of OMAS')

from omas.omas_machine import machine_to_omas, machine_mappings, machine_expression_types, reload_machine_mappings
from omas.omas_utils import *
from omas.omas_physics import cocos_signals

json_opts = dict(indent=1, separators=(',', ': '), sort_keys=True)

OMFITx.TitleGUI('Machine to OMAS mapper GUI')

if 'omas_mappings' not in scratch:
    scratch['omas_mappings'] = {}
scratch['omas_mappings']['user'] = os.environ['USER']


def load_ids():
    if 'omas_mappings' not in OMFIT:
        OMFIT['omas_mappings'] = {}
    OMFIT['omas_mappings']['ids_info'] = omas_info(scratch['omas_mappings']['IDS'])


def load_tree():
    if 'omas_mappings' not in OMFIT:
        OMFIT['omas_mappings'] = {}
    treename = scratch['omas_mappings']['treename'].format(**scratch['omas_mappings']['options'])
    OMFIT['omas_mappings'][treename] = OMFITmds(scratch['omas_mappings']['machine'], treename, scratch['omas_mappings']['pulse'])


def set_machine(location):
    scratch['omas_mappings']['machine'] = tokamak(scratch['omas_mappings']['machine'], 'GPEC')
    reset_confidence()


def set_defaults(location=None):
    scratch['omas_mappings']['location'] = o2u(scratch['omas_mappings']['location'])

    if scratch['omas_mappings']['location'] in mappings:
        scratch['omas_mappings'].update(mappings[scratch['omas_mappings']['location']])
        for exp in machine_expression_types:
            if exp in mappings[scratch['omas_mappings']['location']]:
                scratch['omas_mappings']['expression_type'] = exp
                break
        if 'options' not in scratch['omas_mappings']:
            scratch['omas_mappings']['options'] = {}
        for k in mappings['__options__']:
            if k in scratch['omas_mappings'][scratch['omas_mappings']['expression_type']]:
                scratch['omas_mappings']['options'][k] = mappings['__options__'][k]


def clear_old_expression(location):
    if scratch['omas_mappings']['expression_type'] in scratch['omas_mappings']:
        del scratch['omas_mappings'][scratch['omas_mappings']['expression_type']]
    reset_confidence()


def clear_old_location(location):
    scratch['omas_mappings']['location'] = ''
    reset_confidence()


def reset_confidence(location=None):
    scratch['omas_mappings']['confidence'] = 0


# ================

with OMFITx.same_row():
    OMFITx.Entry(
        "scratch['omas_mappings']['machine']",
        'Device',
        default=tokamak(MainSettings['EXPERIMENT']['device'], 'GPEC'),
        postcommand=set_machine,
        updateGUI=True,
    )
    OMFITx.Entry("scratch['omas_mappings']['pulse']", 'Shot', default=MainSettings['EXPERIMENT']['shot'])
    OMFITx.Button('Reset mappings', reload_machine_mappings, updateGUI=True)
scratch['omas_mappings']['branch'] = ''
# OMFITx.ComboBox("scratch['omas_mappings']['branch']", {'Local': '', 'master': 'master'}, 'GitHub branch', default=None, state='normal')

OMFITx.Separator()

# ================

mappings = machine_mappings(scratch['omas_mappings']['machine'], scratch['omas_mappings']['branch'])

if 'location' not in scratch['omas_mappings']:
    scratch['omas_mappings']['IDS'] = 'equilibrium'
    scratch['omas_mappings']['location'] = 'equilibrium.time_slice.:.global_quantities.ip'
    set_defaults()

mapped_idss = np.unique([k.split('.')[0] for k in mappings.keys()])
m_in_idss = lambda m: '@ ' + m if m in mapped_idss else m
idss = {m_in_idss(k): k for k in list_structures('develop.3')}

with OMFITx.same_row():

    OMFITx.ComboBox(
        "scratch['omas_mappings']['IDS']",
        idss,
        'IDS',
        help=pformat(omas_info_node(scratch['omas_mappings']['IDS'])),
        updateGUI=True,
        state='search',
        postcommand=clear_old_location,
    )
    OMFITx.Button('Load this IDS', load_ids)

# list of possible locations (we hide obsolescent values on purpose!)
m_in_mapping = lambda m: '@ ' + m if m in mappings else m
locations = {}
for k in map(
    lambda x: scratch['omas_mappings']['IDS'] + '.' + o2u(x),
    omas_info(scratch['omas_mappings']['IDS'])[scratch['omas_mappings']['IDS']].flat().keys(),
):
    if omas_info_node(m_in_mapping(k)).get('lifecycle_status', '') != 'obsolescent':
        locations[m_in_mapping(k)] = k

OMFITx.ComboBox(
    "scratch['omas_mappings']['location']",
    locations,
    'IMAS path',
    width=55,
    help=pformat(omas_info_node(scratch['omas_mappings']['location'])),
    updateGUI=True,
    state='search',
    postcommand=set_defaults,
)

OMFITx.ComboBox(
    "scratch['omas_mappings']['expression_type']",
    machine_expression_types,
    'Expression trype',
    updateGUI=True,
    postcommand=clear_old_expression,
)

OMFITx.Entry(
    "scratch['omas_mappings']['%s']" % scratch['omas_mappings']['expression_type'],
    scratch['omas_mappings']['expression_type'],
    default=mappings.get(scratch['omas_mappings']['location'], {}).get(scratch['omas_mappings']['expression_type'], ''),
    postcommand=reset_confidence,
    multiline=True,
)

if scratch['omas_mappings']['location'] in cocos_signals:
    cocos_options = SortedDict()
    cocos_options['?'] = '?'
    for cocos_rule in mappings['__cocos_rules__']:
        if scratch['omas_mappings']['location'] in mappings and re.findall(
            cocos_rule, mappings[scratch['omas_mappings']['location']]['TDI']
        ):
            cocos_options = {'auto': '?'}
            scratch['omas_mappings']['COCOSIO'] = '?'
    cocos_options.update({str(k): k for k in list(range(1, 9)) + list(range(11, 19))})

    OMFITx.ComboBox("scratch['omas_mappings']['COCOSIO']", cocos_options, 'COCOS', default='?', state='normal')
elif 'COCOSIO' in scratch['omas_mappings']:
    del scratch['omas_mappings']['COCOSIO']

if 'TDI' in scratch['omas_mappings']['expression_type']:
    with OMFITx.same_row():
        OMFITx.Entry("scratch['omas_mappings']['treename']", 'treename')
        OMFITx.Button('Load this MDSplus tree', load_tree)

OMFITx.Entry("scratch['omas_mappings']['options']", 'Default options', default={})

# ================
OMFITx.Separator()


def test_mapping():
    user_machine_mappings = get_json_mapping()
    location = 'equilibrium.time_slice.:.global_quantities.beta_normal'
    ods, raw_data = machine_to_omas(
        ODS(),
        scratch['omas_mappings']['machine'],
        scratch['omas_mappings']['pulse'],
        scratch['omas_mappings']['location'],
        options={},
        branch='',
        user_machine_mappings=user_machine_mappings,
    )
    if 'omas_mappings' not in OMFIT:
        OMFIT['omas_mappings'] = {}
    OMFIT['omas_mappings']['raw_data'] = raw_data
    OMFIT['omas_mappings']['ods'] = ods
    data = ods[scratch['omas_mappings']['location']]
    info = omas_info_node(scratch['omas_mappings']['location'])
    printi(f"Successful retrieval of {scratch['omas_mappings']['location']}")
    if hasattr(ods[scratch['omas_mappings']['location']], 'shape'):
        printi(f"{scratch['omas_mappings']['location']} has shape {ods[scratch['omas_mappings']['location']].shape}")
        if len(squeeze(data).shape) == 1:
            plot(squeeze(data))
            title(scratch['omas_mappings']['location'])
            if 'units' in info:
                ylabel(f"[{info['units']}]")
            xlabel('1 st dimension')
        elif len(squeeze(data).shape) == 2:
            cs = contourf(squeeze(data))
            title(scratch['omas_mappings']['location'])
            if 'units' in info:
                colorbar(cs).set_label(f"[{info['units']}]")
            xlabel('2 st dimension')
            ylabel('1 nd dimension')
    else:
        printi(f"Value is: {ods[scratch['omas_mappings']['location']]}")
    pprint(info)


def get_json_mapping():
    tmp = {
        '__options__': scratch['omas_mappings']['options'],
        scratch['omas_mappings']['location']: {
            scratch['omas_mappings']['expression_type']: scratch['omas_mappings'][scratch['omas_mappings']['expression_type']]
        },
    }
    if 'TDI' in scratch['omas_mappings']['expression_type']:
        tmp[scratch['omas_mappings']['location']]['treename'] = scratch['omas_mappings']['treename']
    if 'COCOSIO' in scratch['omas_mappings'] and scratch['omas_mappings']['COCOSIO'] != '?':
        tmp[scratch['omas_mappings']['location']]['COCOSIO'] = scratch['omas_mappings']['COCOSIO']
    return tmp


def print_mapping():
    print('-' * 20)
    printi(json.dumps(get_json_mapping(), **json_opts))


def inject_mapping(update=False):
    from omas.omas_machine import _machine_mappings, _user_machine_mappings

    pprint(get_json_mapping())
    _user_machine_mappings.update(get_json_mapping())
    mappings = machine_mappings(scratch['omas_mappings']['machine'], '')
    if scratch['omas_mappings']['location'] in mappings:
        printi(f"{scratch['omas_mappings']['location']} **TEMPORARILY** added to mapping definitions")
    else:
        raise Exception(f"Failed to add {scratch['omas_mappings']['location']} to mapping definitions")


def update_mapping():
    from omas.omas_machine import update_mapping

    update_mapping(
        machine=scratch['omas_mappings']['machine'],
        location=scratch['omas_mappings']['location'],
        value=get_json_mapping()[scratch['omas_mappings']['location']],
        cocosio=None,
        default_options=scratch['omas_mappings']['options'],
        update_path=False,
    )


with OMFITx.same_row():
    OMFITx.Button('Print', print_mapping)
    OMFITx.Button('Test', test_mapping)
    OMFITx.Button('Inject', inject_mapping)
    OMFITx.Button('Update', update_mapping)


# ================


def suggest_mapping():
    from omas.omas_mongo import get_mongo_credentials
    from pymongo import MongoClient

    server = omas_rcparams['default_mongo_server']
    database = 'omas'
    collection = 'mappings'
    client = MongoClient(server.format(**get_mongo_credentials(server, database, collection)))
    db = client[database]
    coll = db[collection]
    print_mapping()
    res = coll.update(
        {'user': os.environ['USER'], 'location': scratch['omas_mappings']['location']}, copy.copy(scratch['omas_mappings']), upsert=True
    )
    if res['updatedExisting']:
        printw(f'Suggestion updated')
    else:
        printw(f'Suggestion posted')
    # printw(pformat(res))
    return str(res)


OMFITx.Separator()
with OMFITx.same_row():
    OMFITx.Slider("scratch['omas_mappings']['confidence']", [0, 3, 1], 'Confidence level', digits=0, default=1)
    OMFITx.Button('Suggest mapping', suggest_mapping)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii

__all__ = ['OMFITgingred']


class OMFITgingred(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface with GINGRED input files

    :param filename: filename passed to OMFITascii class

    :param \**kw: keyword dictionary passed to OMFITascii class
    """

    def __init__(self, filename, **kw):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.read()
        lines = lines.replace('{', '[')
        lines = lines.replace('}', ']')
        for line in lines.split('\n'):
            line = line.strip().split(';')[0].strip()
            if not len(line):
                continue
            key, value = line.split('=')
            try:
                self[key.strip()] = eval(value.strip())
            except NameError:
                pass

    @dynaSave
    def save(self):
        lines = []
        for key, value in list(self.items()):
            lines.append('%s = %s' % (key, value))
        lines = '\n'.join(lines)
        lines = lines.replace('[', '{')
        lines = lines.replace(']', '}')
        with open(self.filename, 'w') as f:
            f.write(lines)



'''
Provides classes and utility functions for easily using Aurora within OMFIT.
Documentation: https://aurora-fusion.readthedocs.io/
'''

try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise


__all__ = ['OMFITaurora']


class OMFITaurora(SortedDict, OMFITobject):
    r"""
    OMFIT class used to interface with Aurora simulation files.

    :param filename: filename passed to OMFITobject class

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename, namelist=None, geqdsk=None, **kw):

        # AURORA is currently an OPTIONAL dependency, so import here
        from aurora.core import aurora_sim

        OMFITobject.__init__(self, filename, **kw)
        SortedDict.__init__(self)

        # user is initializing class
        if namelist is not None:
            self.dynaLoad = False
        else:
            self.dynaLoad = True

        self.aurora_sim = aurora_sim(namelist=namelist, geqdsk=geqdsk)

    @dynaLoad
    def __getitem__(self, key):
        return getattr(self.aurora_sim, key)

    @dynaLoad
    def __setitem__(self, key, value):
        return setattr(self.aurora_sim, key, value)

    @dynaLoad
    def __getattr__(self, attr):
        return getattr(self.aurora_sim, attr)

    def __setattr__(self, attr, value):
        if 'aurora_sim' in self.__dict__ and attr not in self.__dict__:
            if self.dynaLoad:
                self.load()
                self.dynaLoad = False
            setattr(self.aurora_sim, attr, value)
        else:
            self.__dict__[attr] = value

    @dynaLoad
    def keys(self):
        return self.aurora_sim.__dict__.keys()

    @dynaLoad
    def items(self):
        return self.aurora_sim.__dict__.items()

    @dynaLoad
    def load(self):
        print(f'Loading {self.filename}')
        return self.aurora_sim.load(self.filename)

    @dynaSave
    def save(self):
        print(f'Saving {self.filename}')
        return self.aurora_sim.save(self.filename)



# -*-Python-*-
# Created by eldond 2020-04-07 20:34

"""
This script contains supporting GUIs for efitviewer
"""

from omfit_classes.omfit_efitviewer import (
    omas_features,
    plot_cx_ods,
    load_cx_ods,
    update_cx_time,
    plot_grid,
    dump_plot_command,
    omas_features,
    grab_xylim,
    default_xylim,
    update_overlay,
    add_case,
    delete_case,
    setup_avail_systems,
    setup_special_contours,
    load_contours,
    pick_output_location,
)

from omfit_classes.omfit_efitviewer import update_gui_figure as update_figure

defaultVars(section='co_efv_generic', index=0, location=None, hw_sys=None, out_loc=None)
if out_loc is None:
    output_loc, settings_loc, out, settings, parent_settings, scratch = pick_output_location()
else:
    output_loc = out_loc
    out = eval(out_loc)
    settings_loc = out_loc + "['SETTINGS']"
    settings = eval(settings_loc)


def co_efv_generic():
    """
    Provides an interface to support Customization of Overlays for EFitViewer for generic hardware
    """

    OMFITx.TitleGUI('efitviewer overlay customization ({})'.format(hw_sys))

    # Complain about problems and exit early if there are any
    if hw_sys is None:
        OMFITx.Label('You must specify a hardware system to customize when launching.')
        OMFITx.End()

    special = False
    if hw_sys not in settings['systems'] and hw_sys not in settings['special_systems']:
        OMFITx.Label("{} isn't a recognized hardware system. Perhaps this will work after you load the main efitviewer GUI.".format(hw_sys))
        OMFITx.Label('Valid main options:\n{}'.format('\n'.join(settings['systems'])))
        OMFITx.Label('Valid special options:\n{}'.format('\n'.join(settings['special_systems'])))
        OMFITx.End()
    elif hw_sys not in settings['systems']:
        special = True

    # settings = settings['co_efv'].setdefault(hw_sys, SettingsName())
    ods = out.get('efitviewer_ods', ODS())

    # Try to find the plot method and get its docstring
    po_method_name = 'plot_{}_overlay'.format(hw_sys)
    po_method = getattr(ods, po_method_name, None)
    if po_method is None:
        po_method = getattr(omfit_classes.omfit_efitviewer, po_method_name, None)
    if po_method is None:
        OMFITx.Label('Could not find plot method for {}; unable to display help')
    else:
        settings['co_efv'].setdefault(hw_sys, SettingsName())
        pom_doc_0 = po_method.__doc__
        pom_doc = copy.copy(pom_doc_0)
        p = ':param '
        while p in pom_doc:
            sp = pom_doc.split(p)
            this_par = sp[1]
            if len(sp) > 2:
                pom_doc = p.join([''] + sp[2:])
            else:
                pom_doc = ''
            par_name = this_par.split(':')[0]
            par_description = ':'.join(this_par.split(':')[1:])
            par_description = par_description.split(':return:')[0]
            if par_name in ['ods', 'ax']:
                pass
            elif par_name.startswith('**') or par_name.startswith(r'\**'):
                par_name = par_name.split('**')[1]
                OMFITx.Entry(
                    settings_loc + r"['co_efv']['{}']['pass_in_keywords_{}']".format(hw_sys, par_name),
                    'Additional keywords ({})'.format(par_name),
                    help=par_description,
                    default={},
                    postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                )
                generic_note = (
                    '\n\nThis is a generic setting that is respected by most OMAS plot overlays. It would '
                    'normally be caught in **kw, but its control has been separated out for convenience. '
                    'A redundant specification for this setting in kw will override this one.'
                )
                if not special:
                    OMFITx.Entry(
                        settings_loc + r"['co_efv']['{}']['labelevery']".format(hw_sys),
                        '    labelevery',
                        help=r'Controls which points are labeled. Typically, 0 turns labels off, '
                        r'1 labels every item, 2 labels every item, and so on. ' + generic_note,
                        default=None,
                        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                    )
                    OMFITx.Entry(
                        settings_loc + r"['co_efv']['{}']['notesize']".format(hw_sys),
                        '    notesize',
                        help='Controls size of annotations like point labels' + generic_note,
                        default=None,
                        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                    )
                    OMFITx.Entry(
                        settings_loc + r"['co_efv']['{}']['mask']".format(hw_sys),
                        '    mask',
                        help='Controls which items are displayed. Specify a list of bools that matches the '
                        'number of items that will be plotted.' + generic_note,
                        default=None,
                        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                    )
                    if compare_version(omas.__version__, '0.50.0') >= 0:
                        # These will crash omas plot methods if passed to incompatible versions
                        OMFITx.Entry(
                            settings_loc + r"['co_efv']['{}']['label_ha']".format(hw_sys),
                            '    label_ha',
                            help='Controls horizontal alignment of labels. Should be a string or list of strings. '
                            'Replacing one element with None triggers automatic alignment for that element. '
                            'If the list is too short, newer OMAS versions will be pad it with None (auto). '
                            'The required length is the number of items to label after selection and masking. '
                            'Options: "left", "right", "center".' + generic_note,
                            default=None,
                            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                        )
                        OMFITx.Entry(
                            settings_loc + r"['co_efv']['{}']['label_va']".format(hw_sys),
                            '    label_va',
                            help='Controls vertical alignment of labels. Should be a string or list of strings. '
                            'Replacing one element with None triggers automatic alignment for that element. '
                            'If the list is too short, newer OMAS versions will be pad it with None (auto). '
                            'The required length is the number of items to label after selection and masking. '
                            'Options: "top", "bottom", "center", "baseline", "center_baseline".' + generic_note,
                            default=None,
                            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                        )

                        OMFITx.Entry(
                            settings_loc + r"['co_efv']['{}']['label_r_shift']".format(hw_sys),
                            '    label_r_shift',
                            help='Add offsets to the R coordinates (data units, probably m) of all labels.\n'
                            'Scalar: add the same offset to every label.\n'
                            'Array-like: Allows different offsets for different labels. '
                            'Input will be padded with 0s to meet required length, which is the number of items '
                            'shown (the required length changes if items are removed using selection and masking '
                            'keywords). Older versions of OMAS only support scalars.' + generic_note,
                            default=0.0,
                            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                        )
                        OMFITx.Entry(
                            settings_loc + r"['co_efv']['{}']['label_z_shift']".format(hw_sys),
                            '    label_z_shift',
                            help='Add offsets to the Z coordinates (data units, probably m) of all labels.\n'
                            'Scalar: add the same offset to every label.\n'
                            'Array-like: Allows different offsets for different labels. '
                            'Input will be padded with 0s to meet required length, which is the number of items '
                            'shown (the required length changes if items are removed using selection and masking '
                            'keywords). Older versions of OMAS only support scalars.' + generic_note,
                            default=0.0,
                            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                        )
            elif par_name.startswith('*') or par_name.startswith(r'\*'):
                par_name = par_name.split('*')[1]
                OMFITx.Entry(
                    settings_loc + r"['co_efv']['{}']['pass_in_args_{}']".format(hw_sys, par_name),
                    'Additional args ({})'.format(par_name),
                    help='WARNING: Not ready to support positional arguments. This will not work.\n' + par_description,
                    default=(),
                    postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                    fg='red',
                )
            else:
                OMFITx.Entry(
                    settings_loc + r"['co_efv']['{}']['{}']".format(hw_sys, par_name),
                    par_name,
                    help=par_description,
                    default=None,
                    postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
                )
    return


def co_efv_scaled_boundary():
    """Provides a small GUI for configuring the scaled boundary overlay in EFITviewer"""

    OMFITx.TitleGUI('efitviewer overlay customization (scaled boundary)')

    default_mds_trees = {'DIII-D': 'EFIT01', 'EAST': 'EFIT_EAST'}

    sloc = settings_loc + "['co_efv']['scaled_boundary']"
    sets = settings['co_efv'].setdefault('scaled_boundary', SettingsName())

    OMFITx.Entry(sloc + "['device']", 'Device', default='ITER', postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True))
    sbd = sets['device']
    OMFITx.Entry(
        sloc + "['mds_tree']",
        'MDSplus tree',
        default=default_mds_trees.get(sbd, ''),
        help='may be ignored for future devices',
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    OMFITx.Entry(
        sloc + "['shot']",
        'Shot',
        help='may be ignored for future devices',
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    OMFITx.Entry(
        sloc + "['time']",
        'Time (ms)',
        help='may be ignored for future devices',
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    OMFITx.Entry(
        sloc + "['scale']",
        'Scale factor (pre-offset)',
        default=1 / 3.68 if is_device(sbd, 'ITER') else 1,
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    OMFITx.Entry(sloc + "['x_offset']", 'R offset (m)', default=0, postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True))
    OMFITx.Entry(
        sloc + "['y_offset']",
        'Z offset (m)',
        default=-0.88 / 3.68 if is_device(sbd, 'ITER') else 0,
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    OMFITx.Entry(
        sloc + "['pass_in_keywords_kw']",
        'Plot keywords',
        default={},
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )

    OMFITx.Button('Update', OMFITx.Refresh, updateGUI=True)
    return


def co_efv_custom_script():
    """
    This script creates a GUI for managing options input to a custom overlay script
    """

    OMFITx.TitleGUI('efitviewer overlay customization (custom script)')

    OMFITx.Label('The overlay manager can call the .runNoGUI() method of an OMFITpythonTask instance.')
    OMFITx.Label(
        '''
    The keywords ax and ods will be provided to the .runNoGUI() call
    to pass an Axes instance and an ODS instance. Your script should
    include a defaultVars() call at the top that at minimum accepts
    these keywords. For example: defaultVars(ax=None, ods=None).

    To pass other keywords, enter a dictionary under the
    defaultvars_keywords setting. The complete call is:
    SCRIPT.runNoGUI(ax=ax, ods=ods, **defaultvars_keywords)

    For an example, see the script indicated by the default
    value of "script_loc".'''
    )

    OMFITx.TreeLocationPicker(
        settings_loc + "['co_efv']['custom_script']['script_loc']",
        'script_loc',
        default="OMFIT['EFITtime']['PLOTS']['efitviewer']['example_custom_overlay']",
        help='Location in the OMFIT tree of an OMFITpythonTask instance. Its .runNoGUI() method will be called in '
        'order to generate the overlay. It should have a defaultVars() entry that accepts keywords ax and ods. '
        'It can also accept others that you want to pass using the defaultVars_keywords option..',
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )

    OMFITx.Entry(
        settings_loc + "['co_efv']['custom_script']['defaultvars_keywords']",
        'defaultvars_keywords',
        default={},
        help="Dictionary of keywords and values to pass to the script's defaultVars(), in addition to ods and ax.",
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    return


def co_efv_alt_limiter():
    """
    This script provides a GUI for customizing an alternative limiter overlay
    """

    OMFITx.TitleGUI('efitviewer overlay customization (alternative limiter)')

    sloc = settings_loc + "['co_efv']['alt_limiter']"

    if compare_version(omas.__version__, omas_features['show_wall']) >= 0:
        OMFITx.CheckBox(
            sloc + "['retain_wall']",
            'Also show limiter from ODS',
            default=True,
            postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
        )
    OMFITx.TreeLocationPicker(
        sloc + "['alt_limiter_data_loc']",
        'Alternative limiter data location',
        default="OMFIT['EFITtime']['TEMPLATES']['limiters']['2020-03-23_limiter']['data']",
        help='Limiter data should be an N*2 float array, ' 'where lim[:, 0] gives the R coordinates, and lim[:, 0] gives Z.',
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    OMFITx.Entry(
        sloc + "['alt_limiter_data_array']",
        'Alt limiter data, manual entry',
        default=None,
        help='Limiter data should be a N*2 float array, where lim[:, 0] gives the R coordinates, and lim[:, 1] '
        'gives Z. If data are manually entered here, they will override data from a tree location (above).',
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    OMFITx.Entry(
        sloc + "['scale']",
        'Scale limiter by factor',
        default=1.0,
        help='Use this to convert units to m.',
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )

    OMFITx.Entry(
        sloc + "['pass_in_keywords_kw']",
        'Additional keywords (kw)',
        default={},
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )
    return


def efitviewer_equilibrium_customization():
    """
    Provides a GUI for customizing individual equilibrium cross section plots
    """

    OMFITx.TitleGUI('Equilibrium customization for case {}'.format(index))

    OMFITx.CheckBox(
        settings_loc + "['cases'][{}]['wall']".format(index),
        'Include wall with equilibrium',
        default=True,
        postcommand=lambda location: plot_cx_ods(clear_first=True, gentle=True),
    )

    OMFITx.Entry(
        settings_loc + "['cases'][{}]['eqkw']".format(index),
        'Keywords to pass to plot() & contour(): drawing equilibrium'.format(index),
        default={},
        help="Specify a dictionary like `dict(color='r')`",
        postcommand=update_figure,
    )

    if compare_version(omas.__version__, omas_features['xkw']) >= 0:
        OMFITx.Entry(
            settings_loc + "['cases'][{}]['xkw']".format(index),
            'Keywords to pass to plot(): drawing X-points'.format(index),
            default={'marker': 'x'},
            help='Set marker to "" to disable X-point display',
            postcommand=update_figure,
        )


def efitviewer_contour_picker():
    """GUI for picking contour levels relative to some reference point"""

    OMFITx.TitleGUI('efitviewer: special contour level selection')

    settings.setdefault('custom_contour_spacing', SettingsName())
    cq = settings['contour_quantity']

    OMFITx.Label(
        "Use this feature to support tasks like finding the flux surface that's 1 cm\n"
        "outboard of the midplane, or 3 cm away from the outer strike point."
    )

    OMFITx.ComboBox(
        settings_loc + "['custom_contour_spacing']['quantity']",
        ['R', 'S', 'psi_n', 'psi'],
        'Custom contour spacing in terms of quantity',
        default='R',
        help='R is major radius in m.\n'
        'S is distance along the wall in m.\n'
        'psi_n is normalized poloidal flux.\n'
        'psi is poloidal flux prior to normalization.',
    )

    OMFITx.ComboBox(
        settings_loc + "['custom_contour_spacing']['reference']",
        ['outer_midplane', 'outer_lower_strike', 'outer_upper_strike'],
        'Reference point (ignored for flux quantities like psi)',
        default='outer_midplane',
    )

    OMFITx.Entry(
        settings_loc + "['custom_contour_spacing']['amount']",
        'Contour spacing amount (data units relative to reference)',
        default=[0, 0.1],
        help='The reference point is the LCFS for flux quantities. Otherwise, select the reference above.',
    )

    OMFITx.Button('Load custom {} contours'.format(cq), load_contours, updateGUI=True)
    return


# Call the specified section
try:
    eval(section)()
except NameError:
    if section.startswith('co_efv_'):
        # If a specific customization menu isn't found, fall back to the generic one
        co_efv_generic()
    else:
        raise



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_nc import OMFITnc
from omfit_classes.omfit_eqdsk import OMFITgeqdsk

__all__ = ['OMFITjsolver']


class OMFITjsolver(OMFITnc):
    """
    Class used to parse the ouput of the jsolver equilibirum code
    """

    def to_geqdsk(self, B0, R0, ip, resolution, shot=0, time=0, RBFkw={}):
        """
        maps jsolver solution to a gEQDSK file

        :param B0: scalar vacuum B toroidal at R0

        :param R0: scalar R where B0 is defined

        :param ip: toroidal current

        :param resolution: g-file grid resolution

        :param shot: used to set g-file string

        :param time: used to set g-file string

        :param RBFkw: keywords passed to internal Rbf interpolator

        :return: OMFITgeqdsk object
        """

        return OMFITgeqdsk('g%06d.%05d' % (shot, time)).from_rz(
            self['x']['data'],
            self['z']['data'],
            self['psival']['data'],
            self['p']['data'],
            self['f']['data'],
            self['q']['data'],
            B0,
            R0,
            ip,
            resolution,
            RBFkw=RBFkw,
        )


############################################
if __name__ == '__main__':
    test_classes_main_header()

    j = OMFITjsolver(OMFITsrc + '/../samples/jsolver_eqdsk_11MA_A3_R4.5.cdf')
    g = j.to_geqdsk(8.0, 4.5, 11e6, 65)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii
import numpy as np

__all__ = ['OMFITmmm']


class OMFITmmm(SortedDict, OMFITascii):
    r"""
    OMFIT class used to load from Multi Mode Model output files
    :param filename: filename passed to OMFITascii class
    :param \**kw: keyword dictionary passed to OMFITascii class
    """

    def __init__(self, filename, **kw):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.readlines()[2:]
        # input
        inUnits = lines[0].split()[1:]
        inVars = lines[1].split()[1:]
        inData = []
        i = 2
        for l in lines[i:]:
            if l[0] == '#':
                break
            i += 1
            inData.append([float(d) for d in l.split()])
        # ouput
        outUnits = lines[i + 1].split()[1:]
        outVars = lines[i + 2].split()[1:]
        outData = []
        for l in lines[i + 3 :]:
            outData.append([float(d) for d in l.split()])
        # save data
        for i in range(len(inVars)):
            self[inVars[i]] = np.transpose(inData)[i]
        for i in range(len(outVars)):
            self[outVars[i]] = np.transpose(outData)[i]



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes import namelist

import configobj

__all__ = ['OMFITiniSection', 'OMFITini']


class OMFITiniSection(SortedDict):
    def __init__(self):
        SortedDict.__init__(self, caseInsensitive=True)

    def __delitem__(self, key):
        super().__delitem__(key)
        if '__' + key + '__' in self:
            super().__delitem__('__' + key + '__')

    def _cast(self, data):
        for k in list(data.keys()):
            if isinstance(data[k], dict):
                self[k] = OMFITiniSection()._cast(data[k])
            else:
                self[k] = data[k]
            if k in data.comments and len(data.comments[k]):
                self['__' + k + '__'] = data.comments[k]

        if hasattr(data, 'initial_comment'):
            self.insert(0, '__header__', data.initial_comment)

        return self

    def _uncast(self, data=None, parent=None, depth=None, main=None):
        if data is None:
            main = tmp = configobj.ConfigObj(infile=[], indent_type='    ')
            data = self
            depth = 0
        else:
            tmp = configobj.Section(parent, depth, main)

        for k in self.keys(filter=hide_ptrn, matching=False):
            if isinstance(self[k], OMFITiniSection):
                tmp[k] = self[k]._uncast(self[k], tmp, depth + 1, main)
            else:
                tmp[k] = self[k]

        for k in self.keys(filter=hide_ptrn, matching=True):
            if k != '__header__':
                tmp.comments[k[2:-2]] = self[k]
            else:
                tmp.initial_comment = self[k]

        return tmp

    @staticmethod
    def _translate(self, key):
        """
        This method recursively substitute the ${...} values
        to evaluate the actual string of the expression
        (used for evaluating variables in IPS configuration files)

        :param key: key to be evaluated
        """
        if not isinstance(self[key], str) or '$' not in self[key]:
            return self[key]

        c1 = re.compile(r'\$([^{]\w+)([ /]*)')
        c2 = re.compile(r'\${(\w*)}')
        tmp = self[key]
        while '$' in tmp:
            tmp = re.sub(c1, r'${\1}\2', tmp)
            m = re.search(c2, tmp)
            tmp = tmp[: m.start()] + str(self._ask(m.expand(r'\1'))) + tmp[m.end() :]
        tmp = re.sub(r'\|start\|(.*)\|end\|', r'${\1}', tmp)
        return tmp

    def _ask(self, key):
        if key in self:
            return self[key]
        elif isinstance(self._OMFITparent, OMFITiniSection):
            return self._OMFITparent._ask(key)
        else:
            return '|start|' + key + '|end|'

    def translate(self, key=None):
        if key is not None:
            tmp = self._translate(self, key)
        else:
            tmp = copy.deepcopy(self)
            tmp.walk(self._translate)
        return tmp


class OMFITini(OMFITiniSection, OMFITascii):
    r"""
    OMFIT class used to interface with configuration files (INI files).
    This types of files are used by codes such as IMFIT and IPS

    This class is based on the configobj class https://configobj.readthedocs.io/en/latest/index.html

    :param filename: filename passed to OMFITascii class

    :param \**kw: keyword dictionary passed to OMFITascii class
    """

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self, caseInsensitive=True)
        self.list_values = True
        self.dynaLoad = True

    @staticmethod
    def _interpret(section, key):
        section[key] = namelist.interpreter(section[key], escaped_strings=False)

    @staticmethod
    def _encode(section, key):
        section[key] = namelist.encoder(section[key], escaped_strings=False, dotBoolean=False)

    @dynaLoad
    def load(self):
        """
        Method used to load the content of the file specified in the .filename attribute

        :return: None
        """
        tmp = configobj.ConfigObj(self.filename, list_values=self.list_values)
        tmp.walk(self._interpret)
        self.clear()
        self._cast(tmp)

    @dynaSave
    def save(self):
        """
        Method used to save the content of the object to the file specified in the .filename attribute

        :return: None
        """
        tmp = self._uncast()
        tmp.filename = self.filename
        tmp.walk(self._encode)
        tmp.write()


############################################
if '__main__' == __name__:
    test_classes_main_header()

    tmp = OMFITini(f'{OMFITsrc}/../modules/EPED/eped.config')
    tmp.load()



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

import struct
import numpy as np

__all__ = ['OMFITnimrod']


class OMFITnimrod(SortedDict, OMFITobject):
    r"""
    OMFIT class used to interface with NIMROD output files
    Note: does not support writing

    :param filename: filename passed to OMFITobject class

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename, **kw):
        OMFITobject.__init__(self, filename, **kw)
        SortedDict.__init__(self, caseInsensitive=True)
        self.dynaLoad = True

    contour_labels = [
        'B0_r [T]',  # 1
        'B0_z [T]',  # 2
        'B0_{\\phi} [T]',  # 3
        'J0_r [A/m^2]',  # 4
        'J0_z [A/m^2]',  # 5
        'J0_{\\phi} [A/m^2]',  # 6
        'V0_r [m/s]',  # 7
        'V0_z [m/s]',  # 8
        'V0_{\\phi} [m/s]',  # 9
        'P_0 [Pa]',  # 10
        'P_{0e} [Pa]',  # 11
        'n0 [m^{-3}]',  # 12
        'diff shape',  # 13
        'B_r [T]',  # 14
        'B_z [T]',  # 15
        'B_{\\phi} [T]',  # 16
        'Im B_r [T]',  # 17
        'Im B_z [T]',  # 18
        'Im B_{\\phi} [T]',  # 19
        'J_r [A/m^2]',  # 20
        'J_z [A/m^2]',  # 21
        'J_{\\phi} [A/m^2]',  # 22
        'Im J_r [A/m^2]',  # 23
        'Im J_z [A/m^2]',  # 24
        'Im J_{\\phi} [A/m^2]',  # 25
        'V_r [m/s]',  # 26
        'V_z [m/s]',  # 27
        'V_{\\phi} [m/s]',  # 28
        'Im V_r [m/s]',  # 29
        'Im V_z [m/s]',  # 30
        'Im V_{\\phi} [m/s]',  # 31
        'P [Pa]',  # 32
        'Im P [Pa]',  # 33
        'P_e [Pa]',  # 34
        'Im P_e [Pa]',  # 35
        'n [m^{-3}]',  # 36
        'Im n [m^{-3}]',  # 37
        'conc',  # 38
        'Im conc',  # 39
        'T_e [eV]',  # 40
        'Im T_e [eV]',  # 41
        'T_i [eV]',  # 42
        'Im T_i [eV]',  # 43
        'P_{dilu} [W/m^3]',  # 44
        'Im P_{dilu} [W/m^3]',  # 45
        'P_{brem} [W/m^3]',  # 46
        'Im P_{brem} [W/m^3]',  # 47
        'P_{rec} [W/m^3]',  # 48
        'Im P_{rec} [W/m^3]',  # 49
        'P_{line} [W/m^3]',  # 50
        'Im P_{line} [W/m^3]',  # 51
        'P_{ion} [W/m^3]',  # 52
        'Im P_{ion} [W/m^3]',  # 53
        'n_imp [m^{-3}]',  # 54
        'Im n_imp [m^{-3}]',  # 55
        'n_{ele} [m^{-3}]',  # 56
        'Im n_{ele} [m^{-3}]',  # 57
        'n_{neut} [m^{-3}]',  # 58
        'Im n_{neut} [m^{-3}]',  # 59
        'n_{imp neut} [m^{-3}]',  # 60
    ]
    for iz in range(1, 19):
        contour_labels.append('n_{imp +' + str(iz) + '} [m^{-3}]')
    discharge_labels = [
        'istep',
        'time',
        'divB',
        'totE',  # total energy
        'totIE',  # total internal energy
        'totIEe',  # total internal electron energy
        'totIEi',  # total internal ion energy
        'lnE',  # ln(total energy)
        'lnIE',  # ln(internal energy)
        'grate',  # growth rate
        'Itot',  # current
        'Ipert',  # perturbed current
        'Vloop',  # perturbed loop voltage
        'totflux',  # total flux
        'n0flux',  # n=0 flux
        'bigF',  # F
        'Theta',  # Theta
        'magCFL',  # Eq. Mag-ac CFL
        'NLCFL',  # Nonlinear CFL
        'flowCFL',
    ]  # Flow CFL]

    polflux_labels = ['R', 'Z', 'polflux', 'R*Bphi']

    flsurf_labels = ['t', 'ix', 'polm', 'sqrt(polm)', 'polp', 'sqrt(polp)', 'q_surf', 'par_surf']

    SPIhist_labels = ['thstp', 'thtm', 'ip', 'Rp', 'Zp', 'Pp', 'radius', 'aden', 'taden', 'eden', 'eT', 'ablt', 'tablt', 'asim', 'tsim']

    energy_labels = ['step', 'time', 'imode', 'k', 'E_mag', 'E_kin', 'prad']

    kpraden_labels = [
        'istep',
        'time',
        'qlosd',
        'qloso',
        'qlosb',
        'qlosr',
        'qlosl',
        'qlosi',
        'Nz',
        'Ne',
        'Nz+zi',
        'qlost',
        'elosd',
        'eloso',
        'elosb',
        'elosr',
        'elosl',
        'elosi',
        'elost',
    ]

    kpradnz_labels = ['istep', 'time', 'Nz+2', 'Nz+3', 'Nz+4', 'Nz+5', 'Nz+6', 'Nz+7', 'Nz+8', 'Nz+9', 'Nz+10']

    xy_slice_labels = [
        'ix/mx',  # 0
        'iy/my',  # 1
        'R',  # 2
        'Z',  # 3
        'B0R',  # 4
        'B0Z',  # 5
        'B0Phi',  # 6
        'J0R',  # 7
        'J0Z',  # 8
        'J0Phi',  # 9
        'V0R',  # 10
        'V0Z',  # 11
        'V0Phi',  # 12
        'P0',  # 13
        'PE0',  # 14
        'n0',  # 15
        'diff shape',  # 16
        'BR',  # 17
        'BZ',  # 18
        'BP',  # 19
        'Im BR',  # 20
        'Im BZ',  # 21
        'Im BP',  # 22
        'JR',  # 23
        'JZ',  # 24
        'JP',  # 25
        'Im JR',  # 26
        'Im JZ',  # 27
        'Im JP',  # 28
        'VR',  # 29
        'VZ',  # 30
        'VP',  # 31
        'Im VR',  # 32
        'Im VZ',  # 33
        'Im VP',  # 34
        'P',  # 35
        'Im P',  # 36
        'PE',  # 37
        'Im PE',  # 38
        'n',  # 39
        'Im n',  # 40
        'conc',  # 41
        'Im conc',  # 42
        'Te',  # 43
        'Im Te',  # 44
        'Ti',  # 45
        'Im Ti',
    ]  # 46

    xt_slice_labels = [
        'ix/mx',  # 0
        'istep',  # 1
        't',  # 2
        'R',  # 3
        'Z',  # 4
        'B0R',  # 5
        'B0Z',  # 6
        'B0Phi',  # 7
        'J0R',  # 8
        'J0Z',  # 9
        'J0Phi',  # 10
        'V0R',  # 11
        'V0Z',  # 12
        'V0Phi',  # 13
        'P0',  # 14
        'PE0',  # 15
        'n0',  # 16
        'diff shape',  # 17
        'BR',  # 18
        'BZ',  # 19
        'BP',  # 20
        'Im BR',  # 21
        'Im BZ',  # 22
        'Im BP',  # 23
        'JR',  # 24
        'JZ',  # 25
        'JP',  # 26
        'Im JR',  # 27
        'Im JZ',  # 28
        'Im JP',  # 29
        'VR',  # 30
        'VZ',  # 31
        'VP',  # 32
        'Im VR',  # 33
        'Im VZ',  # 34
        'Im VP',  # 35
        'P',  # 36
        'Im P',  # 37
        'PE',  # 38
        'Im PE',  # 39
        'n',  # 40
        'Im n',  # 41
        'conc',  # 42
        'Im conc',  # 43
        'Te',  # 44
        'Im Te',  # 45
        'Ti',  # 46
        'Im Ti',
    ]  # 47

    xt_slice_kprad_labels = [
        'nimp',  # 48
        'Im nimp',  # 49
        'ne',  # 50
        'Im ne',  # 51
        'qlosd',  # 52
        'Im qlosd',  # 53
        'qloso',  # 54
        'Im qloso',  # 55
        'qlosb',  # 56
        'Im qlosb',  # 57
        'qlosr',  # 58
        'Im qlosr',  # 59
        'qlosl',  # 60
        'Im qlosl',  # 61
        'qlosi',  # 62
        'Im qlosi',  # 63
        'ndn',  # 64
        'Im ndn',
    ]  # 65

    nimhist_labels = [
        'istep',  # 0
        't',  # 1
        'imode',  # 2
        'k',  # 3
        'Br',  # 4
        'Bz',  # 5
        'Bphi',  # 6
        'Im Br',  # 7
        'Im Bz',  # 8
        'Im Bphi',  # 9
        'Jr',  # 10
        'Jz',  # 11
        'Jphi',  # 12
        'Im Jr',  # 13
        'Im Jz',  # 14
        'Im Jphi',  # 15
        'Vr',  # 16
        'Vz',  # 17
        'Vphi',  # 18
        'Im Vr',  # 19
        'Im Vz',  # 20
        'Im Vphi',  # 21
        'P',  # 22
        'Im P',  # 23
        'Pe',  # 24
        'Im Pe',  # 25
        'n',  # 26
        'Im n',  # 27
        'conc',  # 28
        'Im conc',  # 29
        'Te',  # 30
        'Im Te',  # 31
        'Ti',  # 32
        'Im Ti',
    ]  # 33

    def _read_raw(self):
        with open(self.filename, 'rb') as f:
            raw_data = f.read()
        nreals = len(raw_data) // 4
        return np.array(struct.unpack(">" + str(nreals) + 'f', raw_data))

    def _read_table_check(self, l=4):
        with open(self.filename, 'rb') as f:
            d1 = []
            d2 = []
            data = 0
            while data is not None:
                data = self._read_record(f, l)
                if data is not None:
                    if (not len(d2)) or len(data) == len(d2[0]):
                        d2.append(data)
                    elif (not len(d1)) or len(d2) == len(d1[0]):
                        d1.append(np.array(d2))
                        d2 = []
        if not len(d1):
            return np.array(d2)
        return np.array(d1)

    def _read_table(self, l=4):
        with open(self.filename, 'rb') as f:
            d1 = []
            d2 = []
            data = 0
            while data is not None:
                data = self._read_record(f, l)
                if data is not None:
                    if len(data):
                        d2.append(data)
                    else:
                        d1.append(np.array(d2))
                        d2 = []
        if not len(d1):
            return np.array(d2)
        return np.array(d1)

    # Seperate read_table for discharge to fix bug with reading restarts
    def _read_table_dis(self, l=4):
        with open(self.filename, 'rb') as f:
            d1 = []
            data = 0
            while data is not None:
                data = self._read_record(f, l)
                if data is not None and len(data):
                    d1.append(data)

        return np.array(d1)

    def _read_record(self, f, l=4):
        d = f.read(4)
        if not len(d):
            return None
        n = struct.unpack(">1i", d)[0]
        data = np.array(struct.unpack(">" + str(n // l) + 'f', f.read(n)))
        n2 = struct.unpack(">1i", f.read(4))[0]
        if n != n2:
            raise IOError('Error reading record')
        return data

    def _merge_blocks(self, qnt):
        blx = int(self['blx'])
        bly = int(self['bly'])
        nxbl = int(self['nxbl'])
        nybl = int(self['nybl'])

        if len(qnt.shape) == 3:
            tmp = np.zeros(((blx - 1) * nxbl + 1, (bly - 1) * nybl + 1), dtype=qnt.dtype)
        elif len(qnt.shape) == 4:
            tmp = np.zeros(((blx - 1) * nxbl + 1, (bly - 1) * nybl + 1, qnt.shape[2]), dtype=qnt.dtype)

        for ibl in range(nxbl):
            for jbl in range(nybl):
                if len(qnt.shape) == 3:
                    tmp[ibl * (blx - 1) : ibl * (blx - 1) + blx, jbl * (bly - 1) : jbl * (bly - 1) + bly] = qnt[:, :, nybl * ibl + jbl]
                elif len(qnt.shape) == 4:
                    tmp[ibl * (blx - 1) : ibl * (blx - 1) + blx, jbl * (bly - 1) : jbl * (bly - 1) + bly, :] = qnt[
                        :, :, :, nybl * ibl + jbl
                    ]

        return tmp

    def _read_contour(self):
        """
        Parse contour.bin files
        """
        with open(self.filename, 'rb') as f:
            raw_data = f.read()
        bldat = struct.unpack(">" + str(10) + 'i', raw_data[: 4 * 10])
        nbl = self['nbl'] = int(bldat[1])
        nvar = self['nvar'] = int(bldat[3])
        blx = self['blx'] = int(bldat[6] + 1)
        bly = self['bly'] = int(bldat[7] + 1)
        blp = self['blp'] = self['blx'] * self['bly']

        cdata = self._read_raw()

        nslice = self['nslice'] = int((len(cdata) - (11 + nbl * blp * 2 + 6 * (nbl - 1))) / (nbl * (blp + 2) * nvar))
        print('Parsing data: ' + str(nslice) + ' slices')

        r = np.zeros((blx, bly, nbl))
        z = np.zeros((blx, bly, nbl))
        for ibl in range(nbl):
            c = 10 + ibl * (2 * blp + 6) + 2 * blp
            b = 10 + ibl * (2 * blp + 6) + blp
            a = 10 + ibl * (2 * blp + 6)
            r[:, :, ibl] = np.reshape(cdata[a:b], (bly, blx)).T
            z[:, :, ibl] = np.reshape(cdata[b:c], (bly, blx)).T

        nybl = 1
        for ir in range(1, nbl):
            if r[0, 0, 0] != r[0, 0, ir]:
                nybl = ir
                break
        nxbl = nbl // nybl
        self['nxbl'] = nxbl
        self['nybl'] = nybl

        self['z'] = self._merge_blocks(z)
        self['r'] = self._merge_blocks(r)

        cdat2 = cdata[10 + 2 * blp * nbl + 6 * (nbl - 1) :]

        varr = np.zeros((blx, bly, nslice, nbl, nvar))
        for sl in range(nslice):
            for iv in range(nvar):
                for ibl in range(nbl):
                    a = 2 + iv * (blp + 2) + ibl * (blp + 2) * nvar
                    b = 2 + iv * (blp + 2) + ibl * (blp + 2) * nvar + blp
                    varr[:, :, sl, ibl, iv] = np.reshape(cdat2[a:b], (bly, blx)).T
            if sl < nslice:
                cdat2 = cdat2[nbl * nvar * (blp + 2) :]

        printi('Restructuring blocks')
        for k, label in enumerate(self.contour_labels):
            if label[:3] == 'Im ':
                continue
            if k < varr.shape[-1]:
                if 'Im ' + label in self.contour_labels:
                    kI = self.contour_labels.index('Im ' + label)
                    self[label] = self._merge_blocks(varr[:, :, :, :, k] + 1j * varr[:, :, :, :, kI])
                elif ('{imp' in label) and (k + (nvar - 59) // 2 < nvar):
                    kI = k + (nvar - 59) // 2
                    self[label] = self._merge_blocks(varr[:, :, :, :, k] + 1j * varr[:, :, :, :, kI])
                elif k < 59:
                    self[label] = self._merge_blocks(varr[:, :, :, :, k])

    def _plot_contour(self, slice=0, phase=0, nlevels=21):
        """
        Default plot for contour.bin files

        :param slice: slice to plot

        :param phase: plot real part of complex quantities multiplied by exp(1j*phase)

        :param nlevels: number of levels in the contours
        """
        plts = np.array([31, 35, 41, 13, 14, 15, 19, 20, 21, 25, 26, 27])
        ioff()
        try:
            for k, label in enumerate(np.array(self.contour_labels)[plts]):
                if k == 0:
                    ax = pyplot.subplot(3, 4, k + 1)
                else:
                    pyplot.subplot(3, 4, k + 1, sharex=ax, sharey=ax)
                if 'complex' in self[label].dtype.name and phase != 0:
                    contour(self['r'], self['z'], np.real(self[label][:, :, slice] * np.exp(1j * phase)), nlevels)
                else:
                    contour(self['r'], self['z'], self[label][:, :, slice], nlevels)
                pyplot.title('$' + re.sub(r'\(', '\\,(', label) + '$')
                pyplot.gca().set_aspect('equal')
                pyplot.plot(self['r'][-1, :], self['z'][-1, :], 'k')
            autofmt_sharexy()
            pyplot.subplots_adjust(wspace=0.0, hspace=0.2)
            pyplot.suptitle('Slice ' + str(slice))
        finally:
            ion()
            show()

    def _read_polflux(self):
        """
        Parse polflux.bin files
        """
        with open(self.filename, 'rb') as f:
            raw_data = f.read()
        bldat = struct.unpack(">" + str(8) + 'i', raw_data[: 4 * 8])

        mx = self['mx'] = bldat[6] + 1
        my = self['my'] = bldat[7] + 1

        cdata = self._read_table()

        R = self['R'] = np.reshape(cdata[2][: mx * my], (my, mx))
        Z = self['Z'] = np.reshape(cdata[2][mx * my :], (my, mx))
        self['polflux'] = np.ndarray([my, mx, (len(cdata) - 3) // 2])
        self['R*Bphi'] = np.ndarray([my, mx, (len(cdata) - 3) // 2])
        nd = 3
        while nd + 1 < len(cdata):
            polflux = self['polflux'][:, :, (nd - 3) // 2] = np.reshape(cdata[nd][:], (my, mx))
            nd = nd + 1
            RBphi = self['R*Bphi'][:, :, (nd - 4) // 2] = np.reshape(cdata[nd][:], (my, mx))
            nd = nd + 1

    def _plot_polflux(self):
        """
        Default plot for polflux.bin files

        """
        ioff()
        try:
            ax = pyplot.subplot(1, 2, 1)
            contourf(self['R'], self['Z'], self['polflux'])
            pyplot.title('polflux')
            pyplot.subplot(1, 2, 2, sharey=ax)
            contourf(self['R'], self['Z'], self['R*Bphi'])
            pyplot.title('R*Bphi')
            autofmt_sharexy()
            pyplot.subplots_adjust(wspace=0.0, hspace=0.2)
        finally:
            ion()
            show()

    def _read_flsurf(self):
        """
        Parse flsurf.bin files
        """

        cdata = self._read_table()

        for k, label in enumerate(self.flsurf_labels):
            self[label] = cdata[0, :, k]

    def _plot_flsurf(self):
        """
        Default plot for flsurf.bin files

        """
        ioff()
        try:
            pyplot.plot(self['polp'], self['q_surf'], label='q_surf')
            pyplot.plot(self['polp'], self['par_surf'], label='par_surf')
            pyplot.title('q and j|| vs poloidal flux')
            legend()
        finally:
            ion()
            show()

    def _read_discharge(self):
        """
        Parse discharge.bin files
        """
        alld = self._read_table_dis()

        discharge_labels = self.discharge_labels[: alld.shape[-1]]
        for kit, it in enumerate(discharge_labels):
            if it not in ['dummy', 'lnE', 'lnIE']:
                try:
                    self[it] = np.squeeze(alld)[:, kit]
                except IndexError as _excp:
                    raise
                    printe(repr(_excp))

    def _plot_discharge(self):
        """
        Default plot for discharge.bin files
        """
        kit = 0
        for it in self:
            if it not in ['istep', 'time']:
                kit += 1
                if kit == 1:
                    ax = pyplot.subplot(4, 4, kit)
                else:
                    pyplot.subplot(4, 4, kit, sharex=ax)
                pyplot.plot(self['time'] * 1e3, self[it])
                title_inside(it, 0.5, 0.7)
                pyplot.xlabel('time [ms]')
        autofmt_sharex()
        pyplot.suptitle('Discharge summary')

    def _read_kpraden(self):
        """
        Parse kpraden.bin files
        """
        alld = self._read_table()
        kpraden_labels = self.kpraden_labels[: alld.shape[-1]]
        for kit, it in enumerate(kpraden_labels):
            if it not in ['istep']:
                self[it] = np.squeeze(alld)[:, kit]

    def _plot_kpraden(self):
        """
        Default plot for kpraden.bin files
        """
        kit = 0
        for it in self:
            if it not in ['istep', 'time']:
                kit += 1
                if kit == 1:
                    ax = pyplot.subplot(6, 3, kit)
                else:
                    pyplot.subplot(6, 3, kit, sharex=ax)
                pyplot.plot(self['time'] * 1e3, self[it])
                title_inside(it, 0.5, 0.7)
                pyplot.xlabel('time [ms]')
        autofmt_sharex()
        pyplot.suptitle('kprad radiation')

    def _read_kpradnz(self):
        """
        Parse kpradnz.bin files
        """
        alld = self._read_table()
        kpradnz_labels = self.kpradnz_labels[: alld.shape[-1]]
        for kit, it in enumerate(kpradnz_labels):
            if it not in ['istep']:
                self[it] = np.squeeze(alld)[:, kit]

    def _plot_kpradnz(self):
        """
        Default plot for kpradnz.bin files
        """
        kit = 0
        for it in self:
            if it not in ['istep', 'time']:
                kit += 1
                if kit == 1:
                    ax = pyplot.subplot(3, 3, kit)
                else:
                    pyplot.subplot(3, 3, kit, sharex=ax)
                pyplot.plot(self['time'] * 1e3, self[it])
                title_inside(it, 0.5, 0.7)
                pyplot.xlabel('time [ms]')
        autofmt_sharex()
        pyplot.suptitle('kprad impurities')

    def _read_energy(self):
        """
        Parse energy.bin files
        """
        alld = self._read_table_check()
        energy_labels = self.energy_labels[: alld.shape[-1]]
        for kit, it in enumerate(energy_labels):
            if it in ['k', 'imode']:
                continue

            elif it in ['time', 'step', 'prad']:
                self[it] = alld[:, 0, kit]

            else:
                self['modes_k'] = []
                for k in range(alld.shape[1]):
                    kmode = alld[0, k, energy_labels.index('k')]
                    imode = int(alld[0, k, energy_labels.index('imode')])
                    self.setdefault(imode - 1, SortedDict())
                    self[imode - 1][it] = alld[:, imode - 1, kit]
                    self[imode - 1]['keff'] = kmode
                    self['modes_k'].append(kmode)
                # self['modes_k']=np.array(self['modes_k']).astype(int)

    def _plot_energy(self):
        """
        Default plot for energy.bin files
        """
        colors = cm.rainbow(np.linspace(0, 1, len(self['modes_k'])))

        ax1 = pyplot.subplot(2, 1, 1)
        pyplot.ylabel('$E_{\\rm mag}$')
        ax2 = pyplot.subplot(2, 1, 2, sharex=ax1)
        pyplot.ylabel('$E_{\\rm kin}$')

        autofmt_sharex()

        for k in range(len(self['modes_k'])):
            ax1.semilogy(self['time'] * 1e3, self[k]['E_mag'], label=str(self['modes_k'][k]), color=colors[k])

        for k in range(len(self['modes_k'])):
            ax2.semilogy(self['time'] * 1e3, self[k]['E_kin'], label=str(self['modes_k'][k]), color=colors[k])
        ax2.legend(loc=0).draggable(True)

        pyplot.xlim([0, max(self['time'] * 1e3)])
        pyplot.xlabel('time [ms]')

    def _plot_prad(self):
        """
        Prad plot for energy.bin files
        """
        pyplot.semilogy(self['time'] * 1e3, self['prad'], 'k')
        pyplot.ylabel('$P_{\\rm rad}$')
        pyplot.xlim([0, max(self['time'] * 1e3)])
        pyplot.xlabel('time [ms]')

    def _read_grid(self):
        """
        Parse grid.bin files
        """
        alld = self._read_table()
        self['r'] = []
        self['z'] = []
        for k in range(len(alld)):
            self['r'].append(alld[k][:, 0])
            self['z'].append(alld[k][:, 1])

    def _plot_grid(self, alpha=1, linewidth=1, linestyle='-', color=None):
        """
        Default plot for grid.bin files
        """
        pyplot.gca().set_aspect('equal')
        if color is None:
            colors = color_cycle(len(self['r']))
        else:
            colors = len(self['r']) * [color]
        for k in range(len(colors)):
            pyplot.plot(self['r'][k], self['z'][k], color=colors[k], linewidth=linewidth, linestyle=linestyle, alpha=alpha)
        pyplot.xlabel('R [m]')
        pyplot.ylabel('Z [m]')

    def _read_xy_slice(self):
        """
        Parse xy_slice.bin files
        """
        alld = self._read_table()
        nmodes = int((np.size(alld, 2) - 17) // 30)

        for j in range(0, 17):
            self[self.xy_slice_labels[j]] = alld[:, :, j]

        slice = (17, 23, 29, 35, 37, 39, 41, 43, 45, 47)
        for k in range(len(slice) - 1):
            # print(slice[k],slice[k+1],slice[k]-slice[0],slice[k+1]-slice[k])
            for j in range(slice[k], slice[k + 1]):
                label = self.xy_slice_labels[j]
                if label[:3] == 'Im ':
                    continue
                for n in range(nmodes):
                    self.setdefault(n, SortedDict())
                    if 'Im ' + label in self.xy_slice_labels:
                        jI = self.xy_slice_labels.index('Im ' + label)
                        self[n][label] = (
                            alld[:, :, j + (slice[k] - slice[0]) * (nmodes - 1) + n * (slice[k + 1] - slice[k])]
                            + 1j * alld[:, :, jI + (slice[k] - slice[0]) * (nmodes - 1) + n * (slice[k + 1] - slice[k])]
                        )
                    else:
                        self[n][label] = alld[:, :, j + (slice[k] - slice[0]) * (nmodes - 1) + n * (slice[k + 1] - slice[k])]

    def _plot_xy_slice(self, nlevels=21):
        """
        Default plot for xy_slice.bin files
        """
        for k, label in enumerate(self.xy_slice_labels[4:16]):
            pyplot.subplot(3, 4, k + 1)
            pyplot.gca().set_aspect('equal')
            try:
                contour(self['ix/mx'][:, :].T, self['iy/my'][:, :].T, self[label][:, :].T, nlevels)
            except Exception:
                pass
            if k in [8, 9, 10, 11]:
                pyplot.xlabel('ix/mx')
            if k in [0, 4, 8]:
                pyplot.ylabel('iy/my')
            pyplot.title(label)

        autofmt_sharexy()
        pyplot.subplots_adjust(wspace=0.0, hspace=0.2)

    def _read_xt_slice(self):
        """
        Parse xt_slice.bin files
        """
        alld = self._read_table()
        nslice = np.size(alld, 0)

        for n in range(0, nslice):
            self.setdefault(n, SortedDict())
            for j in range(0, 48):
                label = self.xt_slice_labels[j]
                if label[:3] == 'Im ':
                    continue
                if 'Im ' + label in self.xt_slice_labels:
                    jI = self.xt_slice_labels.index('Im ' + label)
                    self[n][label] = alld[n, :, j] + 1j * alld[n, :, jI]
                else:
                    self[n][label] = alld[n, :, j]

            if np.size(alld, 2) >= 59:
                for j in range(48, 66):
                    label = self.xt_slice_kprad_labels[j - 48]
                    if label[:3] == 'Im ':
                        continue
                    if 'Im ' + label in self.xt_slice_kprad_labels:
                        jI = self.xt_slice_kprad_labels.index('Im ' + label)
                        self[n][label] = alld[n, :, j] + 1j * alld[n, :, jI]
                    else:
                        self[n][label] = alld[n, :, j]
            if np.size(alld, 2) >= 61:
                zimp = int((np.size(alld, 2) - 66) / 2)
                for j in range(66, 66 + zimp):
                    label = 'nz+' + str(j - 66)
                    self[n][label] = alld[n, :, j] + 1j * alld[n, :, j + zimp]

    def _plot_xt_slice(self, nlevels=21):
        """
        Default plot for xt_slice.bin files
        """
        colors = color_cycle(len(self))
        for k, label in enumerate(self.xt_slice_labels[i] for i in [18, 20, 24, 26, 30, 32, 36, 40, 44]):
            tlabel = label
            pyplot.subplot(3, 3, k + 1)

            for n in self:
                try:
                    pyplot.plot(self[n]['ix/mx'][:], self[n][label][:], color=colors[n])
                    if k in [0, 2, 4]:
                        pyplot.plot(self[n]['ix/mx'][:], self[n][label[0] + 'Z'][:], linestyle='--', color=colors[n])
                        tlabel = label + ',' + label[0] + 'Z'
                    if k in [6]:
                        pyplot.plot(self[n]['ix/mx'][:], self[n][label[0] + 'E'][:], linestyle='--', color=colors[n])
                        tlabel = label + ',' + label[0] + 'E'
                    if k in [8]:
                        pyplot.plot(self[n]['ix/mx'][:], self[n][label[0] + 'e'][:], linestyle='--', color=colors[n])
                        tlabel = label + ',' + label[0] + 'i'
                except Exception:
                    pass
            pyplot.xlabel('ix/mx')
            pyplot.title(tlabel)
            if k == 1:
                pyplot.gca().legend(self)

        autofmt_sharex()
        pyplot.subplots_adjust(wspace=0.3, hspace=0.2)

    def _read_nimhist(self):
        """
        Parse nimhist.bin files
        """
        alld = self._read_table()
        nmodes = np.size(alld, 1)

        for n in range(0, nmodes):
            self.setdefault(n, SortedDict())
            for j in range(0, 34):
                label = self.nimhist_labels[j]
                if label[:3] == 'Im ':
                    continue
                if 'Im ' + label in self.nimhist_labels:
                    jdrhoI = self.nimhist_labels.index('Im ' + label)
                    self[n][label] = alld[:, n, j] + 1j * alld[:, n, jdrhoI]
                else:
                    self[n][label] = alld[:, n, j]

    def _read_SPIhist(self):
        """
        Parse SPIhist.bin files
        """
        alld = self._read_table()
        SPIhist_labels = self.SPIhist_labels[: alld.shape[-1]]
        for kit, it in enumerate(SPIhist_labels):
            self[it] = alld[:, :, kit]

    def _plot_nimhist(self, nlevels=21):
        """
        Default plot for nimhist.bin files
        """
        colors = cm.rainbow(np.linspace(0, 1, len(self)))
        for k, label in enumerate(self.nimhist_labels[i] for i in [4, 6, 10, 12, 16, 18, 22, 26, 30]):
            tlabel = label
            pyplot.subplot(3, 3, k + 1)

            for n in self:
                try:
                    pyplot.plot(self[n]['t'][:], self[n][label][:], color=colors[n])
                    if k in [0, 2, 4]:
                        pyplot.plot(self[n]['t'][:], self[n][label[0] + 'Z'][:], linestyle='--', color=colors[n])
                        tlabel = label + ',' + label[0] + 'Z'
                    if k in [6]:
                        pyplot.plot(self[n]['t'][:], self[n][label[0] + 'E'][:], linestyle='--', color=colors[n])
                        tlabel = label + ',' + label[0] + 'E'
                    if k in [8]:
                        pyplot.plot(self[n]['t'][:], self[n][label[0] + 'i'][:], linestyle='--', color=colors[n])
                        tlabel = label + ',' + label[0] + 'i'
                except Exception:
                    pass
            pyplot.xlabel('t')
            pyplot.title(tlabel)
            if k == 1:
                pyplot.gca().legend(self).draggable()

        autofmt_sharex()
        pyplot.subplots_adjust(wspace=0.3, hspace=0.2)

    def _read_nimfl(self):
        """
        Parse nimfl.bin files
        """
        alld = self._read_table()
        for k in range(0, len(alld) - 1):
            self[k] = SortedDict()
            self[k]['r'] = alld[k][:, 0]
            self[k]['z'] = alld[k][:, 1]

    def _plot_nimfl(self):
        """
        Default plot for nimfl.bin files
        """
        colors = color_cycle(len(self), cmap_name='hsv')
        limits = {'rmin': 0, 'rmax': 0, 'zmin': 0, 'zmax': 0}
        for k in list(self.keys()):
            pyplot.plot(self[k]['r'], self[k]['z'], '.', color=colors[k])
            limits['zmin'] = min(limits['zmin'], min(self[k]['z']))
            limits['zmax'] = max(limits['zmax'], max(self[k]['z']))
        pyplot.gca().set_aspect('equal')
        pyplot.ylim([limits['zmin'], limits['zmax']])

    @dynaLoad
    def load(self):
        if 'energy' in os.path.split(self.filename)[1]:
            self._read_energy()
        elif 'discharge' in os.path.split(self.filename)[1]:
            self._read_discharge()
        elif 'kpraden' in os.path.split(self.filename)[1]:
            self._read_kpraden()
        elif 'kpradnz' in os.path.split(self.filename)[1]:
            self._read_kpradnz()
        elif 'flsurf' in os.path.split(self.filename)[1]:
            self._read_flsurf()
        elif 'polflux' in os.path.split(self.filename)[1]:
            self._read_polflux()
        elif 'contour' in os.path.split(self.filename)[1]:
            self._read_contour()
        elif 'conrz' in os.path.split(self.filename)[1]:
            self._read_contour()
        elif 'grid' in os.path.split(self.filename)[1]:
            self._read_grid()
        elif 'xy_slice' in os.path.split(self.filename)[1]:
            self._read_xy_slice()
        elif 'xyrz' in os.path.split(self.filename)[1]:
            self._read_xy_slice()
        elif 'xyrz' in os.path.split(self.filename)[1]:
            self._read_xy_slice()
        elif 'xt_slice' in os.path.split(self.filename)[1]:
            self._read_xt_slice()
        elif 'nimfl' in os.path.split(self.filename)[1]:
            self._read_nimfl()
        elif 'nimhist' in os.path.split(self.filename)[1]:
            self._read_nimhist()
        elif 'SPIhist' in os.path.split(self.filename)[1]:
            self._read_SPIhist()
        else:
            self['data'] = self._read_table()

    def plot(self, slice=0, phase=0, nlevels=21, linewidth=1, linestyle='-', color=None, alpha=1):
        """
        Default plot of energy.bin, discharge.bin, contour.bin and _.bin files

        :param slice: (contour.bin) slice to plot

        :param phase: (contour.bin) plot real part of complex quantities multiplied by np.exp(1j*phase)

        :param nlevels: (contour.bin) number of levels in the contours
        """
        if 'energy' in os.path.split(self.filename)[1]:
            self._plot_energy()
        elif 'discharge' in os.path.split(self.filename)[1]:
            self._plot_discharge()
        elif 'kpraden' in os.path.split(self.filename)[1]:
            self._plot_kpraden()
        elif 'kpradnz' in os.path.split(self.filename)[1]:
            self._plot_kpradnz()
        elif 'flsurf' in os.path.split(self.filename)[1]:
            self._plot_flsurf()
        elif 'polflux' in os.path.split(self.filename)[1]:
            self._plot_polflux()
        elif 'contour' in os.path.split(self.filename)[1]:
            self._plot_contour(slice=slice, phase=phase, nlevels=nlevels)
        elif 'conrz' in os.path.split(self.filename)[1]:
            self._plot_contour(slice=slice, phase=phase, nlevels=nlevels)
        elif 'grid' in os.path.split(self.filename)[1]:
            self._plot_grid(linewidth=linewidth, linestyle=linestyle, alpha=alpha, color=color)
        elif 'xy_slice' in os.path.split(self.filename)[1]:
            self._plot_xy_slice(nlevels=nlevels)
        elif 'xyrz' in os.path.split(self.filename)[1]:
            self._plot_xy_slice(nlevels=nlevels)
        elif 'xt_slice' in os.path.split(self.filename)[1]:
            self._plot_xt_slice(nlevels=nlevels)
        elif 'nimfl' in os.path.split(self.filename)[1]:
            self._plot_nimfl()
        elif 'nimhist' in os.path.split(self.filename)[1]:
            self._plot_nimhist()


############################################
if '__main__' == __name__:
    os.chdir(OMFITtmpDir)

    foo = OMFITnimrod(os.path.dirname(__file__) + '/../../samples/nimrod_discharge.bin')
    foo.load()
    print(foo)



# -*-Python-*-
# Created by meneghini at 12 Jul 2019  16:06

defaultVars(prepend="root['INPUTS']['chease.xml']", GUIschema=None, objs=None, debug=False)  # OMFITxml object with xsd file parsed

if objs is None:
    objs = GUIschema[u'xs:schema'][u'xs:element'][u'xs:complexType'][u'xs:all'][u'xs:element']

for obj0 in objs:
    name0 = obj0['@name']
    niceName0 = re.sub('_', ' ', name0)

    # by default hide extra GUI parameters
    if name0.lower() == 'others':
        OMFITx.Tab(niceName0)
        OMFITx.CheckBox("MainScratch['showImasActorsGUIothers']", 'Show extra parameters in this GUI', default=False, updateGUI=True)
        if not MainScratch['showImasActorsGUIothers']:
            continue

    # DOCUMENTATION (help)
    helptxt = ''
    if 'xs:annotation' in obj0 and 'xs:documentation' in obj0['xs:annotation']:
        helptxt = obj0['xs:annotation']['xs:documentation']

    # SIMPLE (Entry)
    if not 'xs:complexType' in obj0:
        if debug:
            print('%s: ENTRY' % name0.ljust(30))
        OMFITx.Entry(prepend + "['%s']" % (name0), niceName0, default=obj0['@default'], help=helptxt)

    # COMPLEX (CompoundGUI)
    elif 'xs:complexType' in obj0 and 'xs:sequence' in obj0['xs:complexType']:
        if 'xs:element' in obj0['xs:complexType']['xs:sequence']:
            if debug:
                print('%s: COMPLEX' % name0.ljust(30))
            OMFITx.Tab(niceName0)
            OMFITx._setDefault(OMFITx._absLocation(prepend + "['%s']" % (name0)), SortedDict())
            OMFITx.CompoundGUI(
                MainScratch['__imasActorGUI__'],
                objs=obj0['xs:complexType']['xs:sequence']['xs:element'],
                prepend=prepend + "['%s']" % (name0),
                debug=debug,
                title='',
            )

    # CHOICE (ComboBox)
    elif 'xs:complexType' in obj0 and 'xs:choice' in obj0['xs:complexType']:
        if debug:
            print('%s: CHOICE' % name0.ljust(30))
        try:
            options = SortedDict()
            for obj1 in tolist(obj0['xs:complexType']['xs:choice']['xs:element']):
                name1 = obj1['@name']
                niceName1 = re.sub('_', ' ', name1)

                options[niceName1] = SortedDict()
                options[niceName1][name1] = SortedDict()

                try:
                    for opt in tolist(obj1[u'xs:complexType'][u'xs:sequence'][u'xs:element']):
                        if '@default' in opt:
                            options[niceName1][name1][opt['@name']] = opt['@default']
                        if '@fixed' in opt:
                            options[niceName1][name1][opt['@name']] = opt['@fixed']
                except KeyError as _excp:
                    printe(treeLocation(obj1)[-1] + ': ' + repr(_excp))
            if len(options):
                OMFITx.ComboBox(prepend + "['%s']" % (name0), options, niceName0, help=helptxt, default=options[options.keys()[0]])
        except KeyError as _excp:
            if debug:
                printe(treeLocation(obj0)[-1] + ': ' + _excp)

    # UNKNOWN (should never happen)
    else:
        print('%s: ??????????' % name0.ljust(30))



# file processed by 2to3
from __future__ import print_function, absolute_import

import sys

if sys.version_info > (3, 0):
    basestring = str

__all__ = ['ddb_float', 'harvest_send', 'harvest_nc']


def ddb_float(float_in):
    """
    Convert float to Decimal compatible with DynamoDB format

    :param float_in: input float

    :return: float in Decimal format
    """
    if not isinstance(float_in, basestring):
        float_in = repr(float(float_in))
    from boto3.dynamodb.types import DYNAMODB_CONTEXT

    return DYNAMODB_CONTEXT.create_decimal(float_in)


def _data_2_message(payload):
    import numpy as np
    import re

    def formatter(data):
        data = '%g' % data
        if '.' in data:
            data = data.lstrip('0')
        data = re.sub(r'e\-0+', 'e-', data)
        data = re.sub(r'e\+0+', 'e+', data)
        return data

    def compress(tmps):
        return tmps
        tmpsc = []
        c = -1
        kold = tmps[0]
        for k in tmps + [tmps[-1] + ' ']:
            if kold == k:
                c += 1
            else:
                if c == 0:
                    tmpsc.append(kold)
                else:
                    tmpsc.append(str(c + 1) + '*' + kold)
                kold = k
                c = 0
        return tmpsc

    message = []
    for what in list(payload.keys()):
        data = payload[what]
        if isinstance(data, (bool, np.bool_)):
            tp = 'b'
            data = str(int(data))
        elif isinstance(data, (list, tuple, np.ndarray)):
            tp = 'a'
            data = re.sub(' ', '', '[' + ','.join(compress(list(map(formatter, np.atleast_1d(data).flatten().tolist())))) + ']')
        elif np.array(data).dtype.kind == 'i':
            tp = 'i'
            data = str(data)
        elif np.array(data).dtype.kind == 'f':
            tp = 'f'
            data = formatter(data)
        elif isinstance(data, basestring):
            tp = 's'
            data = repr(data.strip())
        elif data is None:
            tp = 's'
            data = ''
        else:
            raise Exception('%s objects of type %s are not supported' % (what, type(data)))
        message.append(tp + '@' + what + '=' + data)

    return '|'.join(message)


def harvest_send(payload, table='test_harvest', host=None, port=None, verbose=None, tag=None, protocol=None, process=None):
    """
    Function to send data to the harvesting server

    :param payload: dictionary with payload data

    :param table: table where to put the data

    :param host: harvesting server address
        If None take value from `HARVEST_HOST` environemental variable, or use default `gadb-harvest.duckdns.org` if not set.

    :param port: port the harvesting server is listening on.
        If None take value from `HARVEST_PORT` environemental variable, or use default `0` if not set.

    :param verbose: print harvest message to screen
        If None take value from `HARVEST_VERBOSE` environemental variable, or use default `False` if not set.

    :param tag: tag entry
        If None take value from `HARVEST_TAG` environemental variable, or use default `Null` if not set.

    :param protocol: transmission protocol to be ued (`UDP` or `TCP`)
        If None take value from `HARVEST_PROTOCOL` environemental variable, or use default `UDP` if not set.

    :param process: function passed by user that is called on each of the payload elements prior to submission

    :return: tuple with used (host, port, message)
    """
    import os, socket, copy, random, time

    version = 3

    if host is None:
        if 'HARVEST_HOST' in os.environ:
            host = os.environ['HARVEST_HOST']
        else:
            host = 'gadb-harvest.duckdns.org'

    if protocol is None:
        if 'HARVEST_PROTOCOL' in os.environ:
            protocol = os.environ['HARVEST_PROTOCOL']
        else:
            protocol = 'UDP'

    if port is None:
        if 'HARVEST_PORT' in os.environ:
            port = int(os.environ['HARVEST_PORT'])
        elif protocol == 'UDP':
            port = 32000
        else:
            port = 31000

    if verbose is None:
        if 'HARVEST_VERBOSE' in os.environ:
            verbose = int(os.environ['HARVEST_VERBOSE'])
        else:
            verbose = 0

    for k in payload:
        if '|' in k or '=' in k:
            raise ValueError('The keys of the payload must not contain `|` or `=`:' + k)
    payload_ = payload.__class__()
    if process is None:
        payload_.update(payload_)
    else:
        for item in list(payload.keys()):
            payload_[item] = process(payload[item])

    payload_['_user'] = os.environ['USER']
    payload_['_hostname'] = socket.gethostname()
    payload_['_workdir'] = os.getcwd()

    if '_tag' not in payload_:
        if tag is None:
            if 'HARVEST_TAG' in os.environ:
                tag = os.environ['HARVEST_TAG']
            else:
                tag = ''
        payload_['_tag'] = tag

    message = "%d:%s:%s" % (version, table, _data_2_message(payload_))

    # UDP connection with application level fragmentation
    MTU = 1450
    if protocol in ['UDP']:
        if len(message) < MTU:
            try:
                if protocol == 'UDP':
                    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
                        if sys.version_info > (3, 0):
                            sock.sendto(bytes(message, 'utf8'), (host, port))
                        else:
                            sock.sendto(message, (host, port))
                else:
                    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                        sock.connect((host, port))
                        sock.sendall(message)
                if verbose:
                    print("%s:%d --%s--[%3.3f]-> %s" % (host, port, protocol, len(message) * 1.0 / MTU, message))
            except Exception as _excp:
                if verbose:
                    raise

        else:
            fmt = "&%06d&%03d&%03d&"
            n = MTU - len(fmt % (0, 0, 0))
            split_message = [message[x : x + n] for x in range(0, len(message), n)]
            ID = int((random.randint(0, 10 ** len(str(id(n)))) + id(n)) // 999999)
            for k, message in enumerate(split_message):
                message = (fmt + '%s') % (ID, k, len(split_message), message)
                try:
                    if protocol == 'UDP':
                        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                        sock.sendto(message, (host, port))
                    else:
                        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        sock.connect((host, port))
                        sock.sendall(message)
                        sock.close()
                    time.sleep(0.01)
                    if verbose:
                        print("%s:%d --%s--[%3.3f]-> %s" % (host, port, protocol, len(message) * 1.0 / MTU, message))
                except Exception as _excp:
                    if verbose:
                        print(repr(_excp))

    # TCP connection with fragmentation at TCP layer
    else:
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((host, port))
            #            sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, MTU*256)
            if False:
                sock.sendall(message)
                if verbose:
                    print("%s:%d --%s--> %s" % (host, port, protocol, message))
            else:
                messages = [message[i : i + MTU] for i in range(0, len(message), MTU)]
                while len(messages):
                    # time.sleep(0.01)
                    tmp = messages.pop(0)
                    sock.send(tmp)
                    if verbose:
                        print("%s:%d --%s--> %s" % (host, port, protocol, tmp))
        except Exception as _excp:
            if verbose:
                print(repr(_excp))
        finally:
            try:
                sock.close()
            except Exception:
                pass

    return (host, port, message)


def harvest_nc(filename, entries=None, verbose=False):
    """
    Function that returns data contained in a NetCDF3 file to be sent by harvest_send

    :param filename: NetCDF3 file

    :param entries: subset of variables to loof for in the NetCDF file

    :param verbose: print payload

    :return: payload dictionary
    """
    import os, netCDF4

    payload = {}
    payload['_harvest_filename'] = os.path.abspath(filename)

    nc = netCDF4.Dataset(filename, 'r', format='NETCDF3_CLASSIC')
    if entries is None:
        entries = list(nc.variables.keys())
    for entry in entries:
        if entry in list(nc.variables.keys()):
            try:
                value = nc.variables[entry].getValue()[0]
            except Exception:
                value = nc.variables[entry][:]
            if len(value) == 1:
                payload[entry] = value[0]
            else:
                payload[entry] = value
            if verbose:
                print(str(entry), value[0])
    nc.close()

    return payload


if __name__ == '__main__':
    harvest_send({'test': 'Fail'}, table='test_harvest', verbose=True)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.sortedDict import SortedDict

__all__ = ['OMFITexecutionDiagram']


class OMFITexecutionDiagram(OMFITascii, SortedDict):
    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        l = self.read().splitlines()
        stack = []
        for li in range(len(l)):
            if not l[li].startswith('*'):
                continue
            na = l[li].count('*')
            if len(stack) >= na:
                stack = stack[: na - 1]

            size = l[li].split()[-1].strip('[]')
            size = size.replace('GB', 'e9')
            size = size.replace('MB', 'e6')
            size = size.replace('kB', 'e3')
            size = size.replace('bytes', '')
            size = size.replace('N/A', '0')
            size = float(size)
            time = float(l[li].split()[-2][:-1])
            name = l[li].split()[-3].strip('[]')
            ptr = self
            for k in stack:
                ptr = ptr[k]
            if name in ptr:
                for ii in range(1, 10):
                    if f'{name}_{ii}' not in ptr:
                        name = f'{name}_{ii}'
                        break
            ptr[name] = SortedDict()
            ptr[name].size = size
            ptr[name].time = time
            stack.append(name)

        self.dynaLoad = False

    @dynaLoad
    def print_sorted(self, by='time', nitems=10):
        paths = []
        vals = []

        def traverse(r, p):
            if len(p):
                paths.append(p)
                vals.append(getattr(r, by))

            for k, v in r.items():
                traverse(v, p + [k])

        traverse(self, [])
        pprint(sorted(list(zip(paths, vals)), key=lambda x: x[1])[-nitems:])


if __name__ == '__main__':
    timings = OMFITexecutionDiagram(OMFITsrc + '/../samples/cake_timings_CORI_10_interactive.txt')
    timings.print_sorted(nitems=20)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii

import numpy as np

__all__ = ['OMFITcsv']

_defaults = {
    'delimiter': None,
    'comments': ['#', '!', '%'],
    'fmt': '%.18e',
    'missing_values': ['nan', 'NAN', "N/A", "???"],
    'filling_values': np.nan,
}


class OMFITcsv(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface with Comma Separated Value files

    This class makes use of the np.genfromtxt and np.savetxt methods http://docs.scipy.org/doc/np/user/basics.io.genfromtxt.html

    :param filename: filename passed to OMFITascii class

    :param delimiter: delimiter character that is used to separate the values

    :param comments: charachter or list of characters that will be recognized as comments

    :param fmt: format used to save the data to file

    :param missing_values: The set of strings corresponding to missing data

    :param filling_values: The set of values to be used as default when the data are missing

    :param \**kw: keyword dictionary passed to OMFITascii class
    """

    def __init__(
        self,
        filename,
        delimiter=_defaults['delimiter'],
        comments=_defaults['comments'],
        fmt=_defaults['fmt'],
        missing_values=_defaults['missing_values'],
        filling_values=_defaults['filling_values'],
        **kw,
    ):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        if delimiter == ' ':
            delimiter = None
        for k in _defaults:
            self.OMFITproperties.setdefault(k, eval(k))
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        """
        Method used to load the content of the file specified in the .filename attribute

        :return: None
        """
        with open(self.filename, 'r') as f:
            lines = f.read()

        # try to find out what is the comment used
        # and use pound as the default comment style
        if not isinstance(self.OMFITproperties['comments'], str):
            for comment in self.OMFITproperties['comments']:
                if comment in lines:
                    self.OMFITproperties['comments'] = comment
                    break
        if not isinstance(self.OMFITproperties['comments'], str):
            self.OMFITproperties['comments'] = self.OMFITproperties['comments'][0]

        self['data'] = np.array([])
        try:
            self['data'] = np.genfromtxt(
                self.filename,
                delimiter=self.OMFITproperties['delimiter'],
                comments=self.OMFITproperties['comments'],
                missing_values=self.OMFITproperties['missing_values'],
                filling_values=self.OMFITproperties['filling_values'],
                invalid_raise=True,
            )
        except Exception as _excp:
            try:
                tmpFile = []
                other = []
                nonochar = re.sub(r'[0-9\n\r\+-\. \teEdD' + re.escape(self.OMFITproperties['delimiter']) + ']', '', string.printable)
                with open(self.filename, 'r') as f:
                    for line in f.readlines():
                        if re.search('[' + nonochar + ']', line) is None:
                            tmpFile.append(line)
                        else:
                            other.append(line)

                tmp = OMFITascii('')
                with open(tmp.filename, 'w') as f:
                    f.writelines(tmpFile)

                try:
                    self['data'] = np.genfromtxt(
                        tmp.filename,
                        delimiter=self.OMFITproperties['delimiter'],
                        missing_values=self.OMFITproperties['missing_values'],
                        filling_values=self.OMFITproperties['filling_values'],
                        invalid_raise=True,
                    )

                except Exception:
                    self['data'] = np.genfromtxt(
                        tmp.filename,
                        delimiter=self.OMFITproperties['delimiter'],
                        skip_header=1,
                        missing_values=self.OMFITproperties['missing_values'],
                        filling_values=self.OMFITproperties['filling_values'],
                        invalid_raise=True,
                    )
                    with open(tmp.filename, 'r') as f:
                        header = f.readline()
                    with open(tmp.filename, 'w') as f:
                        f.writelines(header)
                    self['header'] = np.genfromtxt(
                        tmp.filename,
                        delimiter=self.OMFITproperties['delimiter'],
                        missing_values=self.OMFITproperties['missing_values'],
                        filling_values=self.OMFITproperties['filling_values'],
                        invalid_raise=True,
                    )

                self['other'] = other

            except Exception:
                self['data'] = np.genfromtxt(
                    self.filename,
                    delimiter=self.OMFITproperties['delimiter'],
                    comments=self.OMFITproperties['comments'],
                    missing_values=self.OMFITproperties['missing_values'],
                    filling_values=self.OMFITproperties['filling_values'],
                    invalid_raise=False,
                )

        self['data'] = np.atleast_2d(self['data'])
        # if there is no delimiter, then it's a vertical array and data should be transposed
        if (self.OMFITproperties['delimiter'] is None and ' ' not in lines) or (
            self.OMFITproperties['delimiter'] is not None and self.OMFITproperties['delimiter'] not in lines
        ):
            self['data'] = self['data'].T

    @dynaSave
    def save(self):
        """
        Method used to save the content of the object to the file specified in the .filename attribute

        :return: None
        """
        if self.OMFITproperties['delimiter'] is None:
            np.savetxt(self.filename, self['data'], fmt=self.OMFITproperties['fmt'])
        else:
            np.savetxt(self.filename, self['data'], delimiter=self.OMFITproperties['delimiter'], fmt=self.OMFITproperties['fmt'])

    def __save_kw__(self):
        """
        :return: kw dictionary used to save the attributes to be passed when reloading from OMFITsave.txt
        """
        tmp = self.OMFITproperties.copy()
        for k in _defaults:
            if self.OMFITproperties[k] is _defaults[k] or self.OMFITproperties[k] == _defaults[k]:
                tmp.pop(k)
        return tmp



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_base import OMFITtree, OMFITexpression
from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.omfit_path import OMFITpath
from omfit_classes.omfit_namelist import OMFITnamelist
from omfit_classes import OMFITx

__all__ = ['OMFITlatex']


class OMFITlatex(OMFITtree):
    """
    Class used to manage LaTeX projects in OMFIT.

    For testing, please see: samples/sample_tex.tex, samples/sample_bib.bib, samples/placeholder.jpb, and
    samples/bib_style_d3dde.bst .

    :param main: string
        Filename of the main .tex file that will be compiled.

    :param mainroot: string
        Filename without the .tex extension. Leave as None to let this be determined
        automatically. Only use this option if you have a confusing filename that breaks the automatic determinator,
        like blah.tex.tex (hint: try not to give your files confusing names).

    :param local_build: bool
        Deploy pdflatex job locally instead of deploying to remote server specified in SETTINGS

    :param export_path: string
        Your LaTeX project can be quickly exported to this path if it is defined. The project can be
        automatically exported after build if this is defined. Can be defined later in the settings namelist.

    :param export_after_build: bool
        Automatically export after building if export_path is defined. Can be updated later in the settings namelist.

    :param hide_style_files: bool
        Put style files in hidden __sty__ sub tree in OMFIT (they still deploy to top level folder on disk)

    :param debug: bool
        Some data will be saved to OMFIT['scratch']
    """

    def __init__(
        self,
        root=None,
        main='main.tex',
        mainroot=None,
        local_build=True,
        export_path=None,
        export_after_build=False,
        hide_style_files=True,
        debug=False,
        **kw,
    ):
        printd('Initializing omfit_latex instance with main file = {:}'.format(main), topic='latex')
        OMFITtree.__init__(self, **kw)
        self.latex_debug = debug

        # Set up basic definitions and structure
        self.ascii_types = ['tex', 'txt', 'bbl', 'bib', 'aux', 'log', 'bst', 'sh', 'dat']
        self.aux_types = ['bbl', 'aux', 'log', 'out', 'backup', 'blg', 'gz']
        self.style_types = ['cls', 'sty', 'clo', 'bst']
        self.no_load = ['OMFIT_run_command.sh', 'OMFITlatex_settings.dat']
        self.pdflatex_flags = '-halt-on-error'  # This string is inserted between pdflatex and command line arguments
        self.hide_style_files = hide_style_files  # Not in settings because ugly to change after init.
        self.__mainroot = mainroot
        self.__export_path = export_path
        self.__main = main
        self.__export_after_build = export_after_build
        self.__hide_style_files = hide_style_files

        # Define sub-folder to isolate LaTeX stuff from other I/O from the module to avoid accidentally picking up
        # unrelated files
        self.latex_folder = 'OMFITlatex_{}'.format(id(self))
        printd('OMFITlatex self.latex_folder = {}'.format(self.latex_folder), topic='latex')

        # Figure out the parent module root
        self.root = OMFIT if root is None else root

        # Deal with servers an directories
        self.lookup_dirs_and_servers()
        if local_build:
            self.server = 'localhost'
            self.tunnel = ''
            self.remotedir = SERVER['localhost']['workDir']

        # Load settings and sub-folders
        self.prepare_settings()
        printd('Done with OMFITlatex initialization', topic='latex')
        return

    def prepare_settings(self):

        main = self.__main
        mainroot = self.__mainroot
        export_path = self.__export_path
        export_after_build = self.__export_after_build
        hide_style_files = self.__hide_style_files

        # Prepare settings
        try:
            self.setdefault('settings', OMFITnamelist('OMFITlatex_settings.dat'))
        except OSError:
            # This can happen if the class is being initialized while OMFIT is still loading.
            # The project load sequence will put settings in for us, so we can skip this part.
            printd('Unable to prepare OMFITlatex settings at this time.', topic='latex')
            return

        printd('Preparing OMFITlatex settings...', topic='latex')

        # These sub-folders should also be loaded by the project if that's what's happening
        self['__aux__'] = OMFITtree()
        if hide_style_files:
            self['__sty__'] = OMFITtree()

        # Set up name of main file
        if main[-4:] != os.extsep + 'tex':
            self['settings']['main'] = main + os.extsep + 'tex'
        else:
            self['settings']['main'] = main

        # Main file is a path to a file on disk, not just a name: attempt to load existing project.
        if os.sep in self['settings']['main']:
            if os.path.exists(self['settings']['main']) and os.path.isfile(self['settings']['main']):
                # The attempt to load a project found a valid file
                printd('Given initial main file as input...', topic='latex')
                main_dir = os.path.split(self['settings']['main'])[0]
                self['settings']['main'] = os.path.split(self['settings']['main'])[1]
                if export_path is None:
                    export_path = main_dir

                self.grab_project(main_dir=main_dir, get_tex=True)

                printd(
                    'Loaded LaTeX project from main file: {:}, export_path = {:}'.format(self['settings']['main'], export_path),
                    topic='latex',
                )

            else:
                # User attempted to load an existing project, but the file was not valid.
                printw(
                    'Warning: could not load LaTeX project using file {:}. Please check path and try again.'.format(
                        self['settings']['main']
                    )
                )
                self['settings']['export_path'] = os.path.split(self['settings']['main'])[0]
                self['settings']['main'] = os.path.split(self['settings']['main'])[1]  # Reduce to filename w/o path

        # Set up root filename of main file
        if mainroot is None:
            self['settings']['mainroot'] = self['settings']['main'].split(os.extsep + 'tex')[0]
        else:
            self['settings']['mainroot'] = mainroot

        self['settings']['main'] = OMFITexpression("return_variable = parent['mainroot'] + '.tex'")
        self['settings']['outfile'] = OMFITexpression("return_variable = parent['mainroot'] + '.pdf'")
        self['settings']['export_path'] = export_path
        self['settings']['export_after_build'] = export_after_build
        self['settings']['default_build_sequence'] = 'full'
        self['settings']['default_clean_before_build'] = False
        printd('Done preparing OMFITlatex settings.', topic='latex')

    def lookup_dirs_and_servers(self, overwrite=False):
        """
        Looks up workdir, remotedir, tunnel, and server

        Tries to link to module or top-level settings to get information.

        :param overwrite: bool
            Overwrite existing attributes. Otherwise, attributes with valid definitions will not be updated.
        """

        printd('Looking up directory and server information...', topic='latex')
        # Link to module settings and define workdir, etc.
        try:
            module_settings = self.root['SETTINGS']
            _ = module_settings['SETUP']['workDir']
            _ = module_settings['REMOTE_SETUP']
        except (KeyError, TypeError):
            printd('  Drawing dir & server info from MainSettings...', topic='latex')
            ref_workdir = evalExpr(OMFIT['MainSettings']['SETUP']['workDir'])
            ref_remotedir = evalExpr(SERVER['default']['workDir'])
            tunnel = evalExpr(SERVER['default']['tunnel'])
            server = evalExpr(SERVER['default']['server'])
        else:
            printd('  Drawing dir & server info from parent module SETTINGS...', topic='latex')
            ref_workdir = evalExpr(module_settings['SETUP']['workDir'])
            ref_remotedir = evalExpr(module_settings['REMOTE_SETUP']['workDir'])
            tunnel = evalExpr(module_settings['REMOTE_SETUP']['tunnel'])
            server = evalExpr(module_settings['REMOTE_SETUP']['server'])

        printd(
            f'  Found ref_workdir = {repr(ref_workdir)}, ref_remotedir = {repr(ref_remotedir)}, '
            f'server = {repr(server)}, tunnel = {repr(tunnel)}',
            topic='latex',
        )
        # If workdir and remotedir are valid directories, extend them by adding the latex_folder
        if ref_workdir is None or isinstance(ref_workdir, NamelistName):
            # Not sure why it comes back as NamelistName sometimes, but it does.
            workdir = None
        else:
            workdir = ref_workdir + os.sep + self.latex_folder
        if ref_remotedir is None or isinstance(ref_workdir, NamelistName):
            remotedir = None
        else:
            remotedir = ref_remotedir + os.sep + self.latex_folder

        # If workdir and remotedir are missing, signal that we shouldn't trust tunnel and server
        if workdir is None and remotedir is None:
            printd(
                "Failed to find useful info about directories & servers. This can happen during project loading. ",
                topic='latex',
            )
            tunnel = server = None

        # Load data
        if (getattr(self, 'workdir', None) is None) or overwrite:
            self.workdir = workdir
            printd(f'  Set self.workdir = {self.workdir}', topic='latex')
        if (getattr(self, 'remotedir', None) is None) or overwrite:
            self.remotedir = remotedir
            printd(f'  Set self.remotedir = {self.remotedir}', topic='latex')
        if (getattr(self, 'tunnel', None) is None) or overwrite:
            self.tunnel = tunnel
            printd(f'  Set self.tunnel = {self.tunnel}', topic='latex')
        if (getattr(self, 'server', None) is None) or overwrite:
            self.server = server
            printd(f'  Set self.server = {self.server}', topic='latex')

    def __call__(self, **kw):
        return self.run(**kw)

    def grab_project(self, main_dir=None, get_tex=False):
        """
        Grabs project files during init or after build

        :param main_dir: string
            Directory with the files to gather

        :param get_tex: bool
            Gather .tex files as well (only recommended during init)
        """
        if main_dir is None:
            if self.workdir is None:
                self.lookup_dirs_and_servers()
            main_dir = self.workdir
        destination = self
        depth = 0

        def gp_loader(gp_dir, gp_destination=self, gp_depth=0):
            printd('Loading LaTeX project from {:} ...'.format(gp_dir), topic='latex')

            files = glob.glob(gp_dir + os.sep + '*')
            if gp_depth > 20:
                printw('Warning: grab project loader reached maximum recursion depth and aborted.')
                return

            for f in files:
                printd('   Load {}: {} ... '.format(os.path.split(f)[1], f), topic='latex')
                d1 = gp_destination
                if os.path.isdir(f):
                    # Is a directory, load contents
                    printd('    This is a directory (current recursion depth = {:})'.format(gp_depth), topic='latex')
                    f2 = os.path.split(f)[1]
                    gp_loader(gp_dir + os.sep + f2, d1.setdefault(f2, OMFITtree()), gp_depth + 1)
                else:
                    # Not a directory, load file
                    if os.path.splitext(f)[1].strip('.') in self.aux_types:
                        # Hide aux files in separate folder
                        printd('    This is an aux file. (recur depth = {})'.format(gp_depth), topic='latex')
                        d1 = d1.setdefault('__aux__', OMFITtree())
                    elif os.path.splitext(f)[1].strip('.') in self.style_types and self.hide_style_files:
                        # Hide style files in separate folder
                        printd('    This is an style file. (recur depth = {})'.format(gp_depth), topic='latex')
                        d1 = d1.setdefault('__sty__', OMFITtree())
                    elif os.path.split(f)[1] in self.no_load:
                        # Skip certain files
                        printd('    This file should be skipped. (recur depth = {})'.format(gp_depth), topic='latex')
                        continue
                    elif os.path.splitext(f)[1].strip('.') == 'tex':
                        if get_tex:
                            printd(
                                '    This is a tex file, and get_tex=True, so it will be gathered. ' '(recur depth = {})'.format(gp_depth),
                                topic='latex',
                            )
                        else:
                            printd(
                                '    This is a tex file, and get_tex=False, so it will be SKIPPED. ' '(recur depth = {})'.format(gp_depth),
                                topic='latex',
                            )
                            continue

                    else:
                        printd('    This is a regular file (recur depth = {})'.format(gp_depth), topic='latex')

                    try:
                        if os.path.splitext(f)[1].strip('.') in self.ascii_types:
                            d1[os.path.split(f)[1]] = OMFITascii(f)
                        else:
                            d1[os.path.split(f)[1]] = OMFITpath(f)
                    except OMFITexception:
                        printe('Could not load %s' % f)

            if not files:
                printd('    This folder appears to be empty: {:}'.format(gp_dir), topic='latex')

            return  # End of gp_loader()

        gp_loader(main_dir, destination, depth)
        return

    def export(self, export_path=None):
        """
        This is a wrapper for .deploy() that uses the export_path in settings by default.

        :param export_path: string
            The path to which files should be exported.
        """
        # from OMFITx import executable
        if self.workdir is None:
            self.lookup_dirs_and_servers()

        if export_path is None:
            export_path = self['settings']['export_path']

        if export_path is None:
            printe("Cannot export until export_path is defined. Please set self['settings']['export_path']")
        else:
            printi('Exporting OMFITlatex project to {:}...'.format(self['settings']['export_path']))
            printi(
                '   Note: this should happen automatically after building if export_path is set, \n'
                'but you can pass export_path to this function as a keyword to export to \n'
                'somewhere else without changing the default export location. Another \n'
                'advantage is that manual export only copies the essential files and not \n'
                'all the LaTeX aux and log files. You can turn off export_after_build. and \n'
                'just export manually.'
            )

            self.deploy(filename=export_path, updateExistingDir=True)

            # Cleanup hidden files
            command = '\n'.join(
                ['cd {:}'.format(export_path), 'rm -rf __aux__', 'mv __sty__/* .', 'rm -rf __sty__', 'rm -rf __*__', 'rm OMFITsave.txt']
            )
            OMFITx.executable(
                self.root,
                inputs=[],
                outputs=[],
                executable=command,
                clean=False,
                server='localhost',
                tunnel='',
                remotedir=self.workdir,
                workdir=self.workdir,
            )
        return

    def load_sample(self):
        """
        Loads sample LaTeX source files for use as a template, example, or test case
        """
        sample_path = os.path.abspath(os.sep.join([OMFITsrc, '..', 'samples', 'tex', ''])) + os.sep
        self.setdefault('figures', OMFITtree())['placeholder.jpg'] = OMFITpath('{:}figures/placeholder.jpg'.format(sample_path))
        self.setdefault('includes', OMFITtree())['include_me.tex'] = OMFITascii('{:}includes/include_me.tex'.format(sample_path))
        self['includes'].setdefault('subsubfolder_lol', OMFITtree())['subsubsub.tex'] = OMFITascii(
            '{:}includes/subsubfolder_lol/subsubsub.tex'.format(sample_path)
        )
        self['sample_tex.tex'] = OMFITascii('{:}sample_tex.tex'.format(sample_path))
        self['sample_bib.bib'] = OMFITascii('{:}sample_bib.bib'.format(sample_path))
        self['bib_style_d3dde.bst'] = OMFITascii('{:}bib_style_d3dde.bst'.format(sample_path))
        return

    def debug(self):
        """Prints info about the project."""
        len1 = 30  # Short length
        len2 = len1 * 2  # Long length
        form = '{{:{:}}} = {{:}}'.format(len1)
        print('=' * len2)
        print('Debug info for OMFITlatex class in omfit_latex.py')
        print('-' * len2)
        print(form.format('self.__class__.__name__', self.__class__.__name__))
        print(form.format("Main file, self['settings']['main']", self['settings']['main']))
        print(form.format("Expected output file, self['settings']['outfile']", self['settings']['outfile']))
        print(form.format('self.keys()', list(self.keys())))
        print(form.format('self.workdir', self.workdir))
        print('=' * len2)
        return

    def clear_aux(self):
        """Deletes aux files, like .log, .aux, .blg, etc."""
        self['__aux__'] = OMFITtree()
        return

    def build_clean(self):
        """Wrapper for build that calls clear_aux first and then calls build with clean keyword."""
        self.clear_aux()
        self.build(clean_before_build=True, sequence='full')
        return

    def build_pdflatex(self):
        """Wrapper for build that selects the pdflatex sequence."""
        self.build(sequence='pdflatex')
        return

    def build_bibtex(self):
        """Wrapper for build that selects the bibtex sequence."""
        self.build(sequence='bibtex')
        return

    def build_full(self):
        """Wrapper for build that selects the full seqeunce."""
        self.build(sequence='full')
        return

    def check_main_file(self):
        """
        Makes sure the file defined by self['settings']['main'] is present in inputs.
        Backup: check if self['settings']['mainroot'] might work.

        :return: bool
            main file is okay (self['settings']['main'] is present) or might be okay
            (self['settings']['mainroot'] is present).
        """
        if self['settings']['main'] in list(self.keys()):
            printd('Main file is present. Everything is okay.', topic='latex')
            return True
        else:
            printd('Main file not found. Checking for mainroot...', topic='latex')
            filenames = [self[k].filename.split(os.sep)[-1] for k in list(self.keys())]
            rootnames = [os.extsep.join(filename.split(os.extsep)[:-1]) for filename in filenames]
            printd('   File name roots (extensions removed): {:}'.format(rootnames), topic='latex')
            if self['settings']['mainroot'] in rootnames:
                printw('LaTeX warning: You may need to set the name of the main LaTeX file.')
                printd('mainroot was in rootnames, so this might be okay, or you might have a bad day.', topic='latex')
                return True
            else:
                printw('LaTeX warning: Main file name not present in inputs. LaTeX will not run!')
                return False

    def search_for_tex_files(self):
        """
        :return: List of .tex files in self
        """
        filenames = [self[k].filename.split(os.sep)[-1] for k in list(self.keys())]
        printd('  Searching, filenames = {:}'.format(filenames), topic='latex')
        tex_files = [filename for filename in filenames if filename.split(os.extsep)[-1].lower() == 'tex']
        printd('  Search result: tex_files = {:}'.format(tex_files), topic='latex')
        return tex_files

    def build(self, clean_before_build=None, sequence=None):
        """
        :param clean_before_build: bool
            clear temporary workDir before deploying; None pulls value from settings.
        :param sequence: string
            'full', 'pdflatex', 'bibtex', '2pdflatex'; None pulls value from settings.
        """

        printd('Building latex project ({:})...'.format(self['settings']['main']), topic='latex')

        if None in [self.workdir, self.remotedir, self.tunnel, self.server]:
            self.lookup_dirs_and_servers()
        if sequence is None:
            sequence = self['settings']['default_build_sequence']
        if clean_before_build is None:
            clean_before_build = self['settings']['default_clean_before_build']

        pdflatex_outputs = [
            self['settings']['mainroot'] + '.pdf',
            self['settings']['mainroot'] + '.aux',
            self['settings']['mainroot'] + '.log',
            self['settings']['mainroot'] + '.out',
        ]
        bibtex_outputs = [self['settings']['mainroot'] + '.bbl', self['settings']['mainroot'] + '.blg']

        if not self.check_main_file():
            tex_files = self.search_for_tex_files()
            if len(tex_files):
                self['settings']['mainroot'] = '.'.join(tex_files[0].split('.')[:-1])
                print('Main LaTeX file definition was invalid. Main file updated to be: {:}'.format(self['settings']['main']))
                printd("Update main file by changing ['settings']['mainroot']    (no extension)", topic='latex')
            else:
                printw('Could not find any .tex files; cannot automatically update main file.')

        # Define the main commands. Some sequences will repeat these commands.
        mov_command = 'mv %s{,.backup} || true' % self['settings']['outfile']
        bib_command = 'bibtex {:} || true'.format(self['settings']['mainroot'])
        tex_command = 'pdflatex {flags:} {args:}'.format(flags=self.pdflatex_flags, args=self['settings']['mainroot'])

        if sequence == 'pdflatex':
            # Commands for building LaTeX project (just one pass of pdflatex, for minor changes)
            latex_commands = ['echo "OMFIT is trying to build a latex project (one pass pdflatex)..."', mov_command, tex_command]
            outputs = pdflatex_outputs
        elif sequence == '2pdflatex':
            # Commands for building LaTeX project (two passes of pdflatex, for changes that don't affect bibliography)
            latex_commands = [
                'echo "OMFIT is trying to build a latex project (two pass pdflatex)..."',
                mov_command,
                tex_command,
                tex_command,
            ]
            outputs = pdflatex_outputs
        elif sequence == 'bibtex':
            # Commands for building LaTeX project (just bibtex, in case you want to go through the sequence manually)
            latex_commands = ['echo "OMFIT is trying to build a latex project (solo bibtex)..."', mov_command, bib_command]
            outputs = bibtex_outputs
        else:
            # Commands for building LaTeX project (full build) (this is the default if the if/elif statements all fail)
            latex_commands = [
                'echo "OMFIT is trying to build a latex project (FULL BUILD)..."',
                mov_command,
                tex_command,
                bib_command,
                tex_command,
                tex_command,
            ]
            outputs = pdflatex_outputs + bibtex_outputs

        # Make sure outputs exist so that OMFITx.executable doesn't fail if one of them isn't generated
        for output_file in outputs:
            latex_commands = ['touch {}'.format(output_file)] + latex_commands

        latex_commands = ['%s || exit %d' % (v, k + 1) for k, v in enumerate(latex_commands)]
        command = '\n'.join(latex_commands)

        inputs = [self[k] for k in list(self.keys()) if hasattr(self[k], 'filename') and self[k].filename]

        def extend_inputs_from_subfolder(folder, sub, e_inputs, e_outputs, pre='', depth=0):
            if depth > 20:
                printw('Maximum recursion depth reached in extend_inputs_from_subfolder(). Aborting...', topic='latex')
                return e_inputs, e_outputs

            if isinstance(folder[sub], OMFITtree):
                printd('  Checking subfolder {:} ... (recursion depth = {:})'.format(sub, depth), topic='latex')
                i2 = [
                    (
                        folder[sub][kk],
                        '{:}{:}{:}{:}'.format(pre, sub, os.sep, os.path.split(folder[sub][kk].filename)[1])
                        .replace('__aux__', '')
                        .replace('__sty__', ''),
                    )  # Prevent nested __aux__/__aux__/__aux__/...
                    for kk in list(folder[sub].keys())
                    if hasattr(folder[sub][kk], 'filename') and folder[sub][kk].filename
                ]
                e_inputs += i2
                e_outputs += [ii[1].replace('.tex', '.*') for ii in i2 if ii[1].endswith('.tex')]
                printd('   Added more files from subfolder {:}: {:}'.format(sub, i2), topic='latex')
                for kk in list(folder[sub].keys()):
                    e_inputs, e_outputs = extend_inputs_from_subfolder(folder[sub], kk, e_inputs, e_outputs, pre + sub + os.sep, depth + 1)
            return e_inputs, e_outputs

        for k in list(self.keys()):
            if k in ['__aux__', '__sty__']:
                # Special treatment; don't deploy an __aux__ folder but instead dump aux files into main folder.
                # Subfolders in __aux__ will not be supported nor should they exist.
                inputs += [
                    (self[k][m], os.path.basename(self[k][m].filename).replace(k, ''))
                    for m in list(self[k].keys())
                    if hasattr(self[k][m], 'filename') and self[k][m].filename
                ]
            else:
                inputs, outputs = extend_inputs_from_subfolder(self, k, inputs, outputs, '', 0)

        printd('Executing LaTeX commands, \n\n   inputs = \n{:} \n\n   outputs = \n{:}'.format(inputs, outputs), topic='latex')

        if self.latex_debug:
            oscratch = OMFIT['scratch'].setdefault('omfit_latex_test', OMFITtree())
            exa = oscratch.setdefault('executable_arguments', OMFITtree())
            exa['inputs'] = inputs
            exa['outputs'] = outputs
            exa['command'] = command

        OMFITx.executable(
            self.root,
            inputs=inputs,
            outputs=outputs,
            executable=command,
            clean=clean_before_build,
            server=self.server,
            tunnel=self.tunnel,
            remotedir=self.remotedir,
            workdir=self.workdir,
            ignoreReturnCode=False,
        )

        self.grab_project(main_dir=self.workdir)

        printd('Items in OMFITlatex project after build: {:}'.format(list(self.keys())), topic='latex')

        if self['settings']['export_path'] is not None and self['settings']['export_after_build']:
            self.export()
        return

    def run(self):
        """
        Shortcut for opening the output file. If the output file has not been generated yet, a build will be attempted.
        This lets the user easily open the output from the top level context menu for the OMFITlatex instance, which is
        convenient.
        """
        if self['settings']['outfile'] not in list(self.keys()):
            self.build()
        filename = self[self['settings']['outfile']]
        if OMFITaux['GUI'] is not None:
            OMFITaux['GUI'].openFile(thisObject=filename)
        else:
            program = OMFIT['MainSettings']['SETUP']['Extensions'].get('pdf', None)
            if program is not None:
                subprocess.Popen("{} '{}'".format(program, filename), shell=True)
            else:
                printe('Could not identify a program for opening PDF output. You can try to open the file manually:')
                print(filename)
        return

    def open_gui(self):
        OMFIT['scratch']['__latexGUI__'].run(tex=self, singleGUIinstance=False)
        return

    def __popup_menu__(self):
        def run_and_update(command):
            getattr(self, command)()
            OMFITaux['GUI'].update_treeGUI()

        return [
            ['LaTeX>controls...', self.open_gui],
            ['LaTeX>Build (default sequence)', lambda command='build': run_and_update(command)],
            ['LaTeX>Full Build (pdflatex, bibtex, pdflatex, pdflatex)', lambda command='build_full': run_and_update(command)],
            ['LaTeX>Clear aux files', lambda command='clear_aux': run_and_update(command)],
            ['LaTeX>Open PDF', self.run],
            ['LaTeX>Export project', self.export],
        ]



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.omfit_hdf5 import OMFIThdf5
from omfit_classes.omfit_json import OMFITjson

__all__ = [
    'OMFITuedge',
    'uedge_variable',
    'uedge_common_mapper',
    'uedge_common_map_generator',
    'OMFITuedgeBasisInput',
    'OMFITuedgeInput',
    'OMFITuedge',
]


class uedge_variable(str):
    """
    class used to distinguish between strings and uedge variables in the PyUEDGE input deck
    """

    pass


def uedge_common_map_generator(uedge_dir):
    """
    Parse the bbb.v com.v flx.v grd.v files and generate mapper for common blocks

    This is useful because the BASIS version of UEDGE used not
    to require specifying the common blocks for variables.
    Transition to PyUEDGE requires people to add the common block
    information to the old input files.

    To translate old UEDGE files to the new PyUEDGE format use:
    >> OMFITuedgeBasisInput('old_uedge_input.txt').convert()

    :param uedge_dir: folder where UEDGE is installed

    :return: OMFITjson object containing with mapping info
    """
    mapper = OMFITjson('uedge_common_map.json')
    for common in ['aph', 'api', 'bbb', 'com', 'flx', 'grd', 'svr', 'wdf']:
        with open(uedge_dir + '/%s/%s.v' % (common, common), 'r') as f:
            lines = f.read().split('\n')
        block = ''
        var = None
        for line in lines:
            # identify block
            if re.match(r'\*\*\*\*\* \w+:', line):
                block = line.split()[1][:-1]
            # append to previous var comment
            if var is not None and line.strip().startswith('#'):
                mapper[var]['comment'] += '\n' + line
            # parse the line
            tmp = line.strip().split()
            # nothing to do
            if len(tmp) < 2 or tmp[0].startswith('#') or tmp[0].startswith('$'):
                continue
            # if it is a variable then add it to the mapper
            if (
                tmp[1].strip('_') in ['real', 'integer']
                or tmp[1].strip('_').startswith('character')
                or (len(tmp) > 2 and tmp[2].strip('_') in ['real', 'integer'])
                or (len(tmp) > 2 and tmp[2].strip('_').startswith('character'))
            ):
                var = tmp[0].split('(')[0]
                if var not in mapper:
                    mapper[var] = {}
                    mapper[var]['location'] = []
                mapper[var]['var'] = tmp[0]
                mapper[var]['location'].append((common, block))
                mapper[var]['comment'] = line
        mapper.sort()
    return mapper


_uedge_common = []


def uedge_common_mapper(var=None):
    """
    Parses the uedge_common_map.json and caches it for later use

    :param var: either variable name or None

    :return: mapper if var is None else mapping info for var
    """
    import json

    if not len(_uedge_common):
        with open(os.path.split(__file__)[0] + os.sep + 'uedge_common_map.json') as f:
            _uedge_common.append(json.load(f))
    if var is None:
        return _uedge_common[0]
    else:
        var = var.split('(')[0]
        return _uedge_common[0][var]


class OMFITuedgeBasisInput(OMFITascii, SortedDict):
    """
    Class used to parse input files from BASIS version of UEDGE

    To translate old UEDGE files to the new PyUEDGE format use:
    >> OMFITuedgeBasisInput('old_uedge_input.txt').convert()
    """

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        for li in self.read().splitlines():
            if '=' not in li:
                continue
            k, v = li.split('=', 1)
            k = k.strip()
            v = v.split('#')[0].strip()
            try:
                v = ast.literal_eval(v)
            except ValueError:
                v = uedge_variable(v)
            self[k] = v

    def convert(self):
        """
        Convert input files from BASIS version of UEDGE to Python version

        :return: OMFITuedgeInput object
        """
        tmp = OMFITuedgeInput(os.path.basename(self.filename))
        for item in self:
            var = item.split('(')[0]
            if var in uedge_common_mapper():
                mapper = uedge_common_mapper(var)
                common = mapper['location'][0][0]
                mapped_item = item
                if '(' in item:
                    index = np.array(list(map(int, item.split('(')[-1][:-1].split(',')))) - 1
                    mapped_item = var + '[%s]' % ','.join(map(str, index))
                if common not in tmp:
                    tmp[common] = SortedDict()
                tmp[common][mapped_item] = self[item]
        return tmp


class OMFITuedgeInput(OMFITascii, SortedDict):
    """
    Class used to parse input files of Python version of UEDGE
    """

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        for li in self.read().splitlines():
            if '=' not in li:
                continue
            k, v = li.split('=', 1)
            k = k.strip()
            v = v.split('#')[0].strip()

            try:
                if v.startswith('array(') and v.endswith(')'):
                    # Deal with np arrays
                    v = v[6:-1]  # strip the 'array('  and ')'
                    v = np.array(ast.literal_eval(v))  # This can parse high dimension arrays that fromstring() can't
                else:
                    v = ast.literal_eval(v)
            except ValueError:
                v = uedge_variable(v)
            if '.' in k:
                c, k = k.split('.')
                self.setdefault(c, {})
                self[c][k] = v
            else:
                self[k] = v

    @dynaSave
    def save(self):
        tmp = []
        for c in self:
            if isinstance(self[c], dict):
                for k, v in self[c].items():
                    if isinstance(v, uedge_variable):
                        tmp.append(f'{c}.{k}={v}'.replace('\n', '').replace(' ', ''))
                    else:
                        tmp.append(f'{c}.{k}={repr(v)}'.replace('\n', '').replace(' ', ''))
            else:
                v = self[c]
                if isinstance(v, uedge_variable):
                    tmp.append(f'{c}={v}'.replace('\n', '').replace(' ', ''))
                else:
                    tmp.append(f'{c}={repr(v)}'.replace('\n', '').replace(' ', ''))
        with open(self.filename, 'w') as f:
            f.write('\n'.join(tmp))


class OMFITuedge(OMFIThdf5):
    def plot_mesh(self, ax=None, **kw):
        r"""
        Plot mesh

        :param ax: plot axes

        :param \**kw: extra arguments passed to plot
        """
        if ax is None:
            ax = pyplot.gca()

        ix = self['com']['nx']['data']
        iy = self['com']['ny']['data']

        drm = self['com']['rm']['data']
        dzm = self['com']['zm']['data']

        for jy in np.arange(0, iy + 2):
            for jx in np.arange(0, ix + 2):
                ax.plot(drm[jx, jy, [1, 2, 4, 3, 1]], dzm[jx, jy, [1, 2, 4, 3, 1]], **kw)
                if 'color' not in kw:
                    kw['color'] = ax.lines[-1].get_color()

        ax.set_xlabel('$R$ [m]')
        ax.set_ylabel('$Z$ [m]')

        ax.set_aspect('equal', 'datalim')

    def plot_box(self, variable='te', ax=None, levels=None, edgecolor='none', **kw):
        r"""
        Plot UEDGE results for a box mesh

        :param variable: variable to plot

        :param ax: axes

        :param levels: number of levels in contour plot

        :param edgecolor: edgecolor of contour plot

        :param \**kw: extra arguments passed to contour plot
        """
        if ax is None:
            ax = pyplot.gca()

        nice = self.nice_vars(variable)

        comrm = self['com']['rm']['data']
        comzm = self['com']['zm']['data']
        bbbte = self['bbb'][variable]['data']

        yy = bbbte / nice['norm']

        CS = ax.contourf(comzm[:, :, 0], comrm[:, :, 0], yy, levels=levels, **kw)
        ax.contour(comzm[:, :, 0], comrm[:, :, 0], yy, levels=levels, edgecolor=edgecolor)

        ax.set_title('%s [%s]' % (nice['label'], nice['units']))

        ax.set_xlabel('$Z$ [m]')
        ax.set_ylabel('$R$ [m]')

        cbar = pyplot.gcf().colorbar(CS)
        cbar.ax.set_ylabel(nice['units'])

    def plot(self, variable='te', ax=None, edgecolor='none', **kw):
        r"""
        Plot UEDGE results for a poloidal mesh

        :param variable: variable to plot

        :param ax: axes

        :param levels: number of levels in contour plot

        :param edgecolor: edgecolor of contour plot

        :param \**kw: extra arguments passed to PatchCollection
        """

        if ax is None:
            ax = pyplot.gca()

        nice = self.nice_vars(variable)

        from matplotlib.patches import Polygon
        from matplotlib.collections import PatchCollection

        comnx = self['com']['nx']['data']
        comny = self['com']['ny']['data']
        comrm = self['com']['rm']['data']
        comzm = self['com']['zm']['data']

        bbbte = self['bbb'][variable]['data']

        patches = []

        for iy in np.arange(0, comny + 2):
            for ix in np.arange(0, comnx + 2):
                rcol = comrm[ix, iy, [1, 2, 4, 3]]
                zcol = comzm[ix, iy, [1, 2, 4, 3]]
                rcol.shape = (4, 1)
                zcol.shape = (4, 1)
                polygon = Polygon(np.column_stack((rcol, zcol)), True)
                patches.append(polygon)

        vals = np.zeros((comnx + 2) * (comny + 2))

        for iy in np.arange(0, comny + 2):
            for ix in np.arange(0, comnx + 2):
                k = ix + (comnx + 2) * iy
                vals[k] = bbbte[ix, iy] * nice['norm']

        p = PatchCollection(patches, edgecolors=edgecolor, **kw)
        p.set_array(np.array(vals))
        ax.add_collection(p)
        ax.autoscale_view()

        ax.set_title('%s [%s]' % (nice['label'], nice['units']))

        pyplot.xlabel('$R$ [m]')
        pyplot.ylabel('$Z$ [m]')

        ax.set_aspect('equal')

        cbar = pyplot.gcf().colorbar(p)
        cbar.ax.set_ylabel(nice['units'])

    def nice_vars(self, variable):
        """
        return info for a given UEDGE output variable

        :param variable: variable name

        :return: dictionary with units, label, norm
        """
        tmp = {}
        if variable.startswith('t'):
            tmp['label'] = r'T$\mathregular{_%s}$' % variable[1:]
            tmp['units'] = 'eV'
            tmp['norm'] = 1.0 / self['bbb']['ev']['data']
        elif variable.startswith('n'):
            tmp['label'] = r'n$\mathregular{_%s}$' % variable[1:]
            tmp['units'] = '[m^{-3}/1e20]'
            tmp['norm'] = 1.0 / 1.0e20
        elif variable.startswith('u'):
            tmp['label'] = r'V$\mathregular{_%s}$' % variable[1:]
            tmp['units'] = '[(m/s)/1e3]'
            tmp['norm'] = 1.0 / 1e3
        else:
            tmp['label'] = variable
            tmp['units'] = ''
            tmp['norm'] = 1.0
        return tmp
